import numpy as np

from .classifier import Classifier


class NoOp(Classifier):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def score(self, features, axis):
        super().score(features, axis)
        return np.zeros(self.num_labels[axis])

    def update(self, *args, **kwargs):
        super().update(*args, **kwargs)

    def resize(self, *args, **kwargs):
        pass

    def finalize(self, *args, **kwargs):
        super().finalize(*args, **kwargs)
        return self



class Labels:
    def __init__(self, size):
        self.size = size  # Maximum number of labels, NOT enforced here but by the user

    @property
    def all(self):
        raise NotImplementedError()

    @all.setter
    def all(self, labels):
        raise NotImplementedError()

    def save(self, skip=False):
        return (None if skip else self.all), self.size

    def load(self, all_size):
        self.all, self.size = all_size



from collections import deque

from operator import attrgetter
from semstr.util.amr import LABEL_ATTRIB, UNKNOWN_LABEL, LABEL_SEPARATOR
from ucca import core, layer0
from ucca.layer1 import EdgeTags

from ..config import Config


class Node:
    """
    Temporary representation for core.Node with only relevant information for parsing
    """
    def __init__(self, index, swap_index=None, orig_node=None, text=None, paragraph=None, tag=None, label=None,
                 implicit=False, is_root=False, root=None):
        self.index = index  # Index in the configuration's node list
        self.orig_node = orig_node  # Associated core.Node from the original Passage, during training
        self.node_id = orig_node.ID if orig_node else None  # ID of the original node
        self.text = text  # Text for terminals, None for non-terminals
        self.paragraph = paragraph  # int for terminals, None for non-terminals
        self.tag = tag  # Node tag of the original node (Word/Punctuation)
        if label is None:
            self.label = self.category = None
        else:  # Node label prediction is enabled
            self.label, _, self.category = label.partition(LABEL_SEPARATOR)
            if not self.category:
                self.category = None
        # Whether a label has been set yet (necessary because None is a valid label too):
        self.labeled = self.orig_node is not None and self.orig_node.attrib.get(LABEL_ATTRIB) is None
        self.node_index = int(self.node_id.split(core.Node.ID_SEPARATOR)[1]) if orig_node else None
        self.outgoing = []  # Edge list
        self.incoming = []  # Edge list
        self.children = []  # Node list: the children of all edges in outgoing
        self.parents = []  # Node list: the parents of all edges in incoming
        self.outgoing_tags = set()  # String set
        self.incoming_tags = set()  # String set
        self.node = None  # Associated core.Node, when creating final Passage
        self.implicit = implicit  # True or False
        self.swap_index = self.index if swap_index is None else swap_index  # To avoid swapping nodes more than once
        self.height = 0
        self._terminals = None
        self.is_root = is_root
        self.root = root  # Original Passage object this belongs to

    def add_incoming(self, edge):
        self.incoming.append(edge)
        self.parents.append(edge.parent)
        self.incoming_tags.add(edge.tag)

    def add_outgoing(self, edge):
        self.outgoing.append(edge)
        self.children.append(edge.child)
        self.outgoing_tags.add(edge.tag)
        self.height = max(self.height, edge.child.height + 1)
        self._terminals = None  # Invalidate terminals because we might have added some

    @staticmethod
    def attach_nodes(l0, l1, nodes, labeled=True, node_labels=False, verify=False):
        remotes = []  # To be handled after all nodes are created
        linkages = []  # To be handled after all non-linkage nodes are created
        for node in Node.topological_sort(nodes):
            if labeled and verify:
                assert node.text or node.outgoing or node.implicit, "Non-terminal leaf node: %s" % node
            if node.is_linkage:
                linkages.append(node)
            else:
                for edge in node.outgoing:
                    if edge.remote:
                        remotes.append((node, edge))
                    else:
                        edge.child.add_to_l1(l0, l1, node, edge.tag, labeled, node_labels)
        Node.attach_remotes(l1, remotes, verify)
        Node.attach_linkages(l1, linkages, verify)

    @staticmethod
    def topological_sort(nodes):
        """
        Sort self.nodes topologically, each node appearing as early as possible
        Also sort each node's outgoing and incoming edge according to the node order
        """
        levels = {}
        level_by_index = {}
        stack = [node for node in nodes if not node.outgoing]
        while stack:
            node = stack.pop()
            if node.index not in level_by_index:
                parents = [edge.parent for edge in node.incoming]
                if parents:
                    unexplored_parents = [parent for parent in parents
                                          if parent.index not in level_by_index]
                    if unexplored_parents:
                        for parent in unexplored_parents:
                            stack.append(node)
                            stack.append(parent)
                    else:
                        level = 1 + max(level_by_index[parent.index] for parent in parents)
                        levels.setdefault(level, []).append(node)
                        level_by_index[node.index] = level
                else:
                    levels.setdefault(0, []).append(node)
                    level_by_index[node.index] = 0
        nodes = [node for level, level_nodes in sorted(levels.items())
                 for node in sorted(level_nodes, key=lambda x: x.node_index or x.index)]
        for node in nodes:
            node.outgoing.sort(key=lambda x: x.child.node_index or nodes.index(x.child))
            node.incoming.sort(key=lambda x: x.parent.node_index or nodes.index(x.parent))
        return nodes

    def add_to_l1(self, l0, l1, parent, tag, labeled, node_labels):
        """
        Called when creating final Passage to add a new core.Node
        :param l0: Layer0 of the passage
        :param l1: Layer1 of the passage
        :param parent: node
        :param tag: edge tag to link to parent
        :param labeled: there is a reference passage, so keep original node IDs in the "remarks" field
        :param node_labels: whether to add a node label
        """
        edge = self.outgoing[0] if len(self.outgoing) == 1 else None
        if self.text:  # For Word terminals (Punctuation already created by add_punct for parent)
            if parent.node is not None:
                if self.node is None:
                    self.node = parent.node.add(EdgeTags.Terminal, self.get_terminal(l0)).child
                elif self.node not in parent.node.children:
                    parent.node.add(EdgeTags.Terminal, self.node)
        elif edge and edge.child.text and layer0.is_punct(edge.child.get_terminal(l0)):
            if Config().args.verify:
                assert tag == EdgeTags.Punctuation, "Punctuation parent %s's edge tag is %s" % (parent.node_id, tag)
                assert edge.tag == EdgeTags.Terminal, "Punctuation %s's edge tag is %s" % (self.node_id, edge.tag)
            if self.node is None:
                self.node = l1.add_punct(parent.node, edge.child.get_terminal(l0))
                edge.child.node = self.node[0].child
            elif parent.node is not None and self.node not in parent.node.children:
                parent.node.add(EdgeTags.Punctuation, self.node)
        else:  # The usual case
            assert self.node is None, "Trying to create the same node twice (multiple incoming primary edges): " + \
                                      ", ".join(map(str, self.incoming))
            if parent is not None and parent.label and parent.node is None:  # If parent is an orphan and has a a label,
                parent.add_to_l1(l0, l1, None, Config().args.orphan_label, labeled, node_labels)  # link to root
            self.node = l1.add_fnode(None if parent is None else parent.node, tag, implicit=self.implicit)
        if labeled:  # In training
            self.set_node_id()
        if node_labels:
            self.set_node_label()

    @staticmethod
    def attach_remotes(l1, remotes, verify=False):
        for node, edge in remotes:  # Add remote edges
            try:
                assert node.node is not None, "Remote edge from nonexistent node"
                assert edge.child.node is not None, "Remote edge to nonexistent node"
                l1.add_remote(node.node, edge.tag, edge.child.node)
            except AssertionError:
                if verify:
                    raise

    @staticmethod
    def attach_linkages(l1, linkages, verify=False):
        for node in linkages:  # Add linkage nodes and edges
            try:
                link_relation = None
                link_args = []
                for edge in node.outgoing:
                    assert edge.child.node, "Linkage edge to nonexistent node"
                    if edge.tag == EdgeTags.LinkRelation:
                        assert not link_relation, "Multiple link relations: %s, %s" % (link_relation, edge.child.node)
                        link_relation = edge.child.node
                    elif edge.tag == EdgeTags.LinkArgument:
                        link_args.append(edge.child.node)
                    else:
                        Config().log("Ignored non-linkage edge %s from linkage node %s" % (edge, node))
                assert link_relation is not None, "No link relations: %s" % node
                # if len(link_args) < 2:
                #     Config().log("Less than two link arguments for linkage node %s" % node)
                node.node = l1.add_linkage(link_relation, *link_args)
                if node.node_id:  # We are in training and we have a gold passage
                    node.node.extra["remarks"] = node.node_id  # For reference
            except AssertionError:
                if verify:
                    raise

    def get_terminal(self, l0):
        return l0.by_position(self.index)

    def set_node_id(self):
        if self.node is not None and self.node_id is not None:
            self.node.extra["remarks"] = self.node_id  # Keep original node ID for reference

    def set_node_label(self):
        if self.node is not None:
            self.node.attrib[LABEL_ATTRIB] = self.label or UNKNOWN_LABEL

    @property
    def is_linkage(self):
        """
        Is this a LKG type node? (During parsing there are no node types)
        """
        return self.outgoing_tags and self.outgoing_tags.intersection((EdgeTags.LinkRelation, EdgeTags.LinkArgument))

    @property
    def descendants(self):
        """
        Find all children of this node recursively
        """
        result = [self]
        queue = deque(node for node in self.children if node is not self)
        while queue:
            node = queue.popleft()
            if node is not self and node not in result:
                queue.extend(node.children)
                result.append(node)
        return result

    @property
    def terminals(self):
        if self._terminals is None:
            q = [self]
            terminals = []
            while q:
                n = q.pop()
                q.extend(n.children)
                if n.text is not None:
                    terminals.append(n)
            self._terminals = sorted(terminals, key=attrgetter("index"))
        return self._terminals

    @property
    def tok(self):
        return self.orig_node.tok

    def __repr__(self):
        return Node.__name__ + "(" + str(self.index) + \
               ((", " + self.text) if self.text else "") + \
               ((", " + self.node_id) if self.node_id else "") + ")"

    def __str__(self):
        s = '"%s"' % self.text if self.text else self.node_id or str(self.index)
        if self.label:
            s += "/" + self.label
        return s

    def __eq__(self, other):
        return self.index == other.index and self.outgoing == other.outgoing

    def __hash__(self):
        return hash((self.index, tuple(self.outgoing)))

    def __iter__(self):
        return iter(self.outgoing)



from semstr.util.amr import LABEL_ATTRIB, LABEL_SEPARATOR
from ucca import layer1

from .action import Actions
from .config import Config, COMPOUND
from .states.state import InvalidActionError

# Constants for readability, used by Oracle.action
RIGHT = PARENT = NODE = 0
LEFT = CHILD = EDGE = 1
LABEL = 2
ACTIONS = (  # index by [NODE/EDGE][PARENT/CHILD or RIGHT/LEFT][True/False (remote)]
    (  # node actions
        (Actions.Node, Actions.RemoteNode),  # creating a parent
        (Actions.Implicit, None)  # creating a child (remote implicit is not allowed)
    ),
    (  # edge actions
        (Actions.RightEdge, Actions.RightRemote),  # creating a right edge
        (Actions.LeftEdge, Actions.LeftRemote)  # creating a left edge
    )
)


class Oracle:
    """
    Oracle to produce gold transition parses given UCCA passages
    To be used for creating training data for a transition-based UCCA parser
    :param passage gold passage to get the correct edges from
    """
    def __init__(self, passage):
        self.args = Config().args
        self.unlabeled = Config().is_unlabeled()
        l1 = passage.layer(layer1.LAYER_ID)
        self.nodes_remaining = {n.ID for n in l1.all if n is not l1.heads[0] and
                                (self.args.linkage or n.tag != layer1.NodeTags.Linkage) and
                                (self.args.implicit or not is_implicit_node(n))}
        self.edges_remaining = {e for n in passage.nodes.values() for e in n if (self.args.linkage or e.tag not in (
                                layer1.EdgeTags.LinkRelation, layer1.EdgeTags.LinkArgument)) and
                                (self.args.implicit or not is_implicit_node(e.child)) and
                                (self.args.remote or not is_remote_edge(e))}
        if self.unlabeled:  # Keep only one edge between each pair of nodes, since we cannot distinguish between them
            unique_edges = {(e.parent.ID, e.child.ID, is_remote_edge(e)): e for e in self.edges_remaining}
            self.edges_remaining = set(unique_edges.values())
        self.passage = passage
        self.found = False
        self.log = None

    def get_actions(self, state, all_actions, create=True):
        """
        Determine all zero-cost action according to current state
        Asserts that the returned action is valid before returning
        :param state: current State of the parser
        :param all_actions: Actions object used to map actions to IDs
        :param create: whether to create new actions if they do not exist yet
        :return: dict of action ID to Action
        """
        actions = {}
        invalid = []
        for action in self.generate_actions(state):
            all_actions.generate_id(action, create=create)
            if action.id is not None:
                try:
                    if self.args.validate_oracle:
                        state.check_valid_action(action, message=True)
                    actions[action.id] = action
                except InvalidActionError as e:
                    invalid.append((action, e))
        if self.args.validate_oracle:
            assert actions, self.generate_log(invalid, state)
        return actions

    def generate_log(self, invalid, state):
        self.log = "\n".join(["Oracle found no valid action",
                              state.str("\n"), self.str("\n"),
                              "Actions returned by the oracle:"] +
                             ["  %s: %s" % (action, e) for (action, e) in invalid] or ["None"])
        return self.log

    def generate_actions(self, state):
        """
        Determine all zero-cost action according to current state
        :param state: current State of the parser
        :return: generator of Action items to perform
        """
        self.found = False
        if state.stack:
            s0 = state.stack[-1]
            incoming, outgoing = [[e for e in l if e in self.edges_remaining]
                                  for l in (s0.orig_node.incoming, s0.orig_node.outgoing)]
            if not incoming and not outgoing and not self.need_label(s0):
                yield self.action(Actions.Reduce)
            else:
                # Check for node label action: if all terminals have already been connected
                if self.need_label(s0) and not any(is_terminal_edge(e) for e in outgoing):
                    yield self.action(s0, LABEL, 1)

                # Check for actions to create new nodes
                for edge in incoming:
                    if edge.parent.ID in self.nodes_remaining and not is_implicit_node(edge.parent) and (
                                not is_remote_edge(edge) or
                                # Allow remote parent if all its children are remote/implicit
                                all(is_remote_edge(e) or is_implicit_node(e.child) for e in edge.parent)):
                        yield self.action(edge, NODE, PARENT)  # Node or RemoteNode

                for edge in outgoing:
                    if edge.child.ID in self.nodes_remaining and is_implicit_node(edge.child) and (
                            not is_remote_edge(edge)):  # Allow implicit child if it is not remote
                        yield self.action(edge, NODE, CHILD)  # Implicit

                if len(state.stack) > 1:
                    s1 = state.stack[-2]
                    # Check for node label action: if all terminals have already been connected
                    if self.need_label(s1) and not any(is_terminal_edge(e) for e in
                                                       self.edges_remaining.intersection(s1.orig_node.outgoing)):
                        yield self.action(s1, LABEL, 2)

                    # Check for actions to create binary edges
                    for edge in incoming:
                        if edge.parent.ID == s1.node_id:
                            yield self.action(edge, EDGE, RIGHT)  # RightEdge or RightRemote

                    for edge in outgoing:
                        if edge.child.ID == s1.node_id:
                            yield self.action(edge, EDGE, LEFT)  # LeftEdge or LeftRemote
                        elif state.buffer and edge.child.ID == state.buffer[0].node_id and \
                                len(state.buffer[0].orig_node.incoming) == 1:
                            yield self.action(Actions.Shift)  # Special case to allow discarding simple children quickly

                    if not self.found:
                        # Check if a swap is necessary, and how far (if compound swap is enabled)
                        related = dict([(edge.child.ID,  edge) for edge in outgoing] +
                                       [(edge.parent.ID, edge) for edge in incoming])
                        distance = None  # Swap distance (how many nodes in the stack to swap)
                        for i, s in enumerate(state.stack[-3::-1], start=1):  # Skip top two: checked above, not related
                            edge = related.pop(s.node_id, None)
                            if edge is not None:
                                if not self.args.swap:  # We have no chance to reach it, so stop trying
                                    self.remove(edge)
                                    continue
                                if distance is None and self.args.swap == COMPOUND:  # Save the first one
                                    distance = min(i, Config().args.max_swap)  # Do not swap more than allowed
                                if not related:  # All related nodes are in the stack
                                    yield self.action(Actions.Swap(distance))
                                    break

        if not self.found:
            yield self.action(Actions.Shift if state.buffer else Actions.Finish)

    def action(self, edge, kind=None, direction=None):
        self.found = True
        if kind is None:
            return edge  # Will be just an Action object in this case
        if kind == LABEL:
            return Actions.Label(direction, orig_node=edge.orig_node, oracle=self)
        node = (edge.parent, edge.child)[direction] if kind == NODE else None
        tag = "" if self.unlabeled else edge.tag
        return ACTIONS[kind][direction][is_remote_edge(edge)](tag=tag, orig_edge=edge, orig_node=node, oracle=self)

    def remove(self, edge, node=None):
        self.edges_remaining.discard(edge)
        if node is not None:
            self.nodes_remaining.discard(node.ID)

    def need_label(self, node):
        return self.args.node_labels and not self.args.use_gold_node_labels \
               and not node.labeled and node.orig_node.attrib.get(LABEL_ATTRIB)

    def get_label(self, state, node):
        true_label = raw_true_label = None
        if node.orig_node is not None:
            raw_true_label = node.orig_node.attrib.get(LABEL_ATTRIB)
        if raw_true_label is not None:
            true_label, _, _ = raw_true_label.partition(LABEL_SEPARATOR)
            if self.args.validate_oracle:
                try:
                    state.check_valid_label(true_label, message=True)
                except InvalidActionError as e:
                    raise InvalidActionError("True label is invalid: " + "\n".join(map(str, (true_label, state, e))))
        return true_label, raw_true_label

    def str(self, sep):
        return "nodes left: [%s]%sedges left: [%s]" % (" ".join(self.nodes_remaining), sep,
                                                       " ".join(map(str, self.edges_remaining)))

    def __str__(self):
        return str(" ")


def is_terminal_edge(edge):
    return edge.tag == layer1.EdgeTags.Terminal


def is_remote_edge(edge):
    return edge.attrib.get("remote", False)


def is_implicit_node(node):
    return node.attrib.get("implicit", False)



from collections import OrderedDict

from ..model_util import load_json, save_json


class Classifier:
    """
    Interface for classifier used by the parser.
    """

    def __init__(self, config, labels, input_params=None):
        """
        :param config: Config to get hyperparameters from
        :param labels: dict of axis (string) -> Labels object, can be updated later to add new axes and labels
        :param input_params: dict of feature type name -> FeatureInformation
        """
        self.config = config
        self.labels = labels
        self.input_params = input_params
        self.model_type = self.config.args.classifier
        self.learning_rate = self.config.args.learning_rate
        self.learning_rate_decay = self.config.args.learning_rate_decay
        self.model = self.labels_t = None
        self.is_frozen = False
        self.updates = self.epoch = self.best_score = 0
        self._num_labels = self.num_labels

    @property
    def num_labels(self):
        return OrderedDict((a, len(l.all)) for a, l in self.labels.items()) if self.labels \
            else OrderedDict((a, len(l) if l else s) for a, (l, s) in self.labels_t.items()) if self.labels_t else {}
    
    @property
    def input_dim(self):
        raise NotImplementedError()

    def score(self, features, axis):
        if not self.is_frozen:
            self._update_num_labels()

    def init_features(self, features, axes, train=False, passage=None, lang=None):
        pass

    def update(self, features, axis, pred, true, importance=None):
        """
        Update classifier weights according to predicted and true labels
        :param features: extracted feature values
        :param axis: axis of the label we are predicting
        :param pred: label predicted by the classifier (non-negative integer bounded by num_labels[axis])
        :param true: true labels (non-negative integers bounded by num_labels[axis])
        :param importance: how much to scale the update for the weight update for each true label
        """
        assert not self.is_frozen, "Cannot update a frozen model"
        self._update_num_labels()

    def _update_num_labels(self):
        """
        self.num_labels is a property, and so updated automatically when a label is added to self.labels,
        but we may need to resize the weight matrices whenever that happens
        """
        if self._num_labels != self.num_labels:
            self._num_labels = self.num_labels
            self.resize()

    def resize(self):
        raise NotImplementedError()

    def finalize(self, *args, finished_epoch=False, **kwargs):
        assert not self.is_frozen, "Cannot freeze a frozen model"
        self._update_num_labels()
        if finished_epoch:
            self.epoch += 1

    def finished_step(self, train=False):
        """
        Called by the parser when a single step is finished
        """
        pass

    def finished_item(self, train=False, renew=True):
        """
        Called by the parser when a whole item is finished
        """
        pass

    def transition(self, action, axis):
        pass

    def save(self, filename, skip_labels=(), **kwargs):
        """
        Save all parameters to file
        """
        d = OrderedDict((
            ("type", self.model_type),
            ("axes", OrderedDict(
                (a, OrderedDict((("index", i), ("labels", l.save(skip=a in skip_labels)))))  # (all, size)
                for i, (a, l) in enumerate(self.labels.items())
            )),
            ("is_frozen", self.is_frozen),
            ("learning_rate", self.learning_rate),
            ("learning_rate_decay", self.learning_rate_decay),
            ("updates", self.updates),
            ("epoch", self.epoch),
            ("best_score", self.best_score),
        ) + tuple(kwargs.items()))
        self.save_model(filename, d)
        save_json(filename + ".json", d)

    def save_model(self, filename, d):
        """
        Save all parameters to file
        """
        pass

    def load(self, filename):
        """
        Load all parameters from file
        """
        d = self.load_file(filename, clear=True)
        model_type = d.get("type")
        assert model_type is None or model_type == self.model_type, "Model type does not match: %s" % model_type
        self.labels_t = OrderedDict((a, l["labels"]) for a, l in sorted(d["axes"].items(), key=lambda x: x[1]["index"]))
        self.is_frozen = d["is_frozen"]
        self.config.args.learning_rate = self.learning_rate = d["learning_rate"]
        self.config.args.learning_rate_decay = self.learning_rate_decay = d["learning_rate_decay"]
        self.updates = d.get("updates", 0)
        self.epoch = d.get("epoch", 0)
        self.best_score = d.get("best_score", 0)
        self.load_model(filename, d)

    def load_model(self, filename, d):
        pass

    @classmethod
    def get_property(cls, filename, prop):
        return cls.load_file(filename).get(prop, None)

    LOADED = {}  # Cache for loaded JSON files

    @classmethod
    def load_file(cls, filename, clear=False):
        d = cls.LOADED.get(filename)
        if d is None:
            d = load_json(filename + ".json")
            cls.LOADED[filename] = d
        if clear:
            cls.LOADED.clear()
        return d

    def all_params(self):
        return OrderedDict((k, v.all) for k, v in self.labels.items())

    def __str__(self):
        return "Labels: %s, features: %s" % tuple(map(dict_value, (self.num_labels, self.input_dim)))

    def print_params(self, max_rows=10):
        pass


def dict_value(d):
    return next(iter(d.values())) if len(d) == 1 else ", ".join("%s: %d" % i for i in d.items())



import numpy as np


def randomize_orthonormal(*parameters, activation=None):  # Saxe et al., 2014 (https://arxiv.org/abs/1312.6120)
    for param in parameters:
        shape = param.shape()
        if len(shape) == 2 and shape[0] == shape[1] > 1:
            init, _, _ = np.linalg.svd(np.random.randn(*shape))
            if str(activation) == "relu":
                init *= np.sqrt(2)
            param.set_value(init)



from itertools import combinations

import numpy as np
import os
import pytest
from functools import partial
from glob import glob
from numpy.testing import assert_allclose, assert_array_equal
from semstr.convert import FROM_FORMAT
from semstr.util.amr import WIKIFIER
from ucca import ioutil

from tupa.config import Config

FORMATS = ("ucca", "amr", "conllu", "sdp")


def pytest_addoption(parser):
    parser.addoption("--write-oracle-actions", action="store_true", help="overwrite oracle action expected results")
    parser.addoption("--write-features", action="store_true", help="overwrite feature expected results")
    parser.addoption("--multitask", action="store_true", help="test multitask parsing")


@pytest.fixture
def write_oracle_actions(request):
    return request.config.getoption("--write-oracle-actions")


@pytest.fixture
def write_features(request):
    return request.config.getoption("--write-features")


def pytest_generate_tests(metafunc):
    if "formats" in metafunc.fixturenames:
        formats = [[f] for f in FORMATS]
        if metafunc.config.getoption("--multitask"):
            formats += [[FORMATS[0], f] for f in FORMATS[1:]]
        metafunc.parametrize("formats", formats, ids="-".join)


def create_config():
    c = Config("", "-m", "test")
    c.update({"verbose": 2, "timeout": 1, "embedding_layer_dim": 1, "ner_dim": 1, "action_dim": 1, "lemma_dim": 1,
              "max_words_external": 3, "word_dim_external": 1, "word_dim": 1, "max_words": 3, "max_lemmas": 3,
              "max_tags": 3, "max_pos": 3, "max_deps": 3, "max_edge_labels": 3, "max_puncts": 3, "max_action_types": 3,
              "max_ner_types": 3, "edge_label_dim": 1, "tag_dim": 1, "pos_dim": 1, "dep_dim": 1, "optimizer": "sgd",
              "output_dim": 1, "layer_dim": 2, "layers": 3, "lstm_layer_dim": 2, "lstm_layers": 3,
              "max_action_ratio": 10, "update_word_vectors": False, "copy_shared": None})
    c.update_hyperparams(shared={"lstm_layer_dim": 2, "lstm_layers": 1}, ucca={"word_dim": 2},
                         amr={"max_node_labels": 3, "max_node_categories": 3,
                              "node_label_dim": 1, "node_category_dim": 1})
    return c


@pytest.fixture
def config():
    return create_config()


@pytest.fixture
def empty_features_config():
    c = create_config()
    c.update({"ner_dim": 0, "action_dim": 0, "word_dim_external": 0, "word_dim": 0, "lemma_dim": 0, "node_label_dim": 0,
              "node_category_dim": 0, "edge_label_dim": 0, "tag_dim": 0, "pos_dim": 0, "dep_dim": 0})
    return c


class Settings:
    SETTINGS = ("implicit", "linkage", "unlabeled")
    VALUES = {"unlabeled": (None, [])}
    INCOMPATIBLE = (("linkage", "unlabeled"),)

    def __init__(self, *args):
        for attr in self.SETTINGS:
            setattr(self, attr, attr in args)

    @classmethod
    def all(cls):
        return [Settings(*c) for n in range(len(cls.SETTINGS) + 1) for c in combinations(cls.SETTINGS, n)
                if not any(all(s in c for s in i) for i in cls.INCOMPATIBLE)]

    def dict(self):
        return {attr: self.VALUES.get(attr, (False, True))[getattr(self, attr)] for attr in self.SETTINGS}

    def list(self):
        return [attr for attr in self.SETTINGS if getattr(self, attr)]

    def suffix(self):
        return "_".join([""] + self.list())

    def __str__(self):
        return "-".join(self.list()) or "default"


@pytest.fixture
def default_setting():
    return Settings()


def passage_files(*formats):
    return [f for fo in formats or ["*"] for f in glob("test_files/*." + ("xml" if fo == "ucca" else fo))
            if not f.endswith(".txt")]


def load_passage(filename, annotate=False):
    WIKIFIER.enabled = False
    converters = {k: partial(c, annotate=annotate) for k, c in FROM_FORMAT.items()}
    passages = ioutil.read_files_and_dirs(filename, converters=converters, attempts=1, delay=0)
    try:
        return next(iter(passages))
    except StopIteration:
        return passages


def basename(filename):
    return os.path.basename(os.path.splitext(filename)[0])


@pytest.fixture
def test_passage():
    return next(iter(ioutil.read_files_and_dirs(("test_files/120.xml",))))


def weight_decay(model):
    try:
        return np.float_power(1 - model.classifier.weight_decay, model.classifier.updates)
    except AttributeError:
        return 1


def assert_all_params_equal(*params, decay=1):
    for key, param in sorted(params[0].items()):
        for p in params[1:]:
            try:
                assert_allclose(decay * param, p[key], rtol=1e-6)
            except TypeError:
                assert_array_equal(param, p[key])


def remove_existing(filename):
    for f in glob(filename + ".*"):
        os.remove(f)



import dynet as dy
import numpy as np

from .constants import ACTIVATIONS, INITIALIZERS, RNNS, CategoricalParameter
from .mlp import MultilayerPerceptron
from .sub_model import SubModel
from .util import randomize_orthonormal
from ...model_util import MISSING_VALUE


class BiRNN(SubModel):
    def __init__(self, config, args, model, **kwargs):
        super().__init__(**kwargs)
        self.config = config
        self.args = args
        self.model = model
        self.dropout = self.args.dropout
        self.lstm_layers = self.args.lstm_layers
        self.lstm_layer_dim = self.args.lstm_layer_dim
        self.embedding_layers = self.args.embedding_layers if self.lstm_layer_dim and self.lstm_layers else 0
        self.embedding_layer_dim = self.args.embedding_layer_dim
        self.max_length = self.args.max_length
        self.activation = CategoricalParameter(ACTIVATIONS, self.args.activation)
        self.init = CategoricalParameter(INITIALIZERS, self.args.init)
        self.rnn_builder = CategoricalParameter(RNNS, self.args.rnn)
        self.input_reps = self.empty_rep = self.indexed_dim = self.indexed_num = None
        self.mlp = MultilayerPerceptron(self.config, self.args, self.model, params=self.params,
                                        layers=self.embedding_layers, layer_dim=self.embedding_layer_dim,
                                        output_dim=self.lstm_layer_dim)

    def init_params(self, indexed_dim, indexed_num):
        """
        Initialize BiRNN builder
        :return: total output dimension of BiRNN
        """
        if self.lstm_layer_dim and self.lstm_layers:
            if self.params:
                assert self.indexed_dim == indexed_dim, "Input dim changed: %d != %d" % (self.indexed_dim, indexed_dim)
                assert self.indexed_num == indexed_num, "Input num changed: %d != %d" % (self.indexed_num, indexed_num)
            else:
                self.indexed_dim = indexed_dim
                self.indexed_num = indexed_num
                randomize_orthonormal(*self.init_rnn_params(indexed_dim), activation=self.activation)
                self.config.print("Initializing BiRNN: %s" % self, level=4)

    def init_rnn_params(self, indexed_dim):
        rnn = dy.BiRNNBuilder(self.lstm_layers, self.lstm_layer_dim if self.embedding_layers else indexed_dim,
                              self.lstm_layer_dim, self.model, self.rnn_builder())
        self.params["birnn"] = rnn
        return [p for f, b in rnn.builder_layers for r in (f, b) for l in r.get_parameters() for p in l]

    def init_features(self, embeddings, train=False):
        """
        Set the value of self.input_reps (and self.empty_rep) given embeddings for the whole input sequence
        :param embeddings: list of [(key, list of vectors embeddings per time step)] per feature
        :param train: are we training now?
        """
        if self.params:
            keys, embeddings = zip(*embeddings)
            inputs = [self.mlp.evaluate(zip(keys, es), train=train) for es in zip(*embeddings)]  # join each time step
            self.config.print("Transducing %d inputs with dropout %s" %
                              (len(inputs), self.dropout if train else "disabled"), level=4)
            self.input_reps = self.transduce(inputs, train)
            expected = min(len(inputs), self.max_length or np.iinfo(int).max)
            assert len(self.input_reps) == expected, \
                "transduce() returned incorrect number of elements: %d != %d" % (len(self.input_reps), expected)
            self.empty_rep = dy.inputVector(np.zeros(self.lstm_layer_dim, dtype=float))

    def transduce(self, inputs, train):
        birnn = self.params["birnn"]
        if train:
            birnn.set_dropout(self.dropout)
        else:
            birnn.disable_dropout()
        return birnn.transduce(inputs[:self.max_length])

    def evaluate(self, indices):
        """
        :param indices: indices of inputs
        :return: list of BiRNN outputs at given indices
        """
        if self.params:
            assert len(indices) == self.indexed_num, "Input size mismatch: %d != %d" % (len(indices), self.indexed_num)
            return [("/".join(self.save_path),
                     self.empty_rep if i == MISSING_VALUE else self.get_representation(i)) for i in indices]
        return []

    def get_representation(self, i):
        return self.input_reps[min(i, self.max_length - 1)]

    def transition(self, action):
        pass

    def save_sub_model(self, d, *args):
        values = super().save_sub_model(
            d,
            ("lstm_layers", self.lstm_layers),
            ("lstm_layer_dim", self.lstm_layer_dim),
            ("embedding_layers", self.embedding_layers),
            ("embedding_layer_dim", self.embedding_layer_dim),
            ("max_length", self.max_length),
            ("activation", str(self.activation)),
            ("init", str(self.init)),
            ("dropout", self.dropout),
            ("rnn", str(self.rnn_builder)),
            ("indexed_dim", self.indexed_dim),
            ("indexed_num", self.indexed_num),
            ("input_dim", self.mlp.input_dim),
            ("gated", self.mlp.gated),
        ) if self.lstm_layer_dim and self.lstm_layers else []
        self.config.print("Saving BiRNN: %s" % self, level=4)
        return values

    def load_sub_model(self, d, *args, **kwargs):
        d = super().load_sub_model(d, *args, **kwargs)
        if d:
            self.args.lstm_layers = self.lstm_layers = d["lstm_layers"]
            self.args.lstm_layer_dim = self.mlp.output_dim = self.lstm_layer_dim = d["lstm_layer_dim"]
            self.args.embedding_layers = self.mlp.layers = self.embedding_layers = d["embedding_layers"]
            self.args.embedding_layer_dim = self.mlp.layer_dim = self.embedding_layer_dim = d["embedding_layer_dim"]
            self.args.max_length = self.max_length = d["max_length"]
            self.args.rnn = self.rnn_builder.string = d["rnn"]
            self.args.activation = self.mlp.activation.string = self.activation.string = d["activation"]
            self.args.init = self.mlp.init.string = self.init.string = d["init"]
            self.args.dropout = self.mlp.dropout = self.dropout = d["dropout"]
            self.indexed_dim = self.mlp.input_dim = d["indexed_dim"]
            self.indexed_num = d["indexed_num"]
            self.mlp.input_dim = d.get("input_dim")
            self.mlp.gated = d.get("gated")
        else:
            self.args.lstm_layers = self.lstm_layers = self.args.lstm_layer_dim = self.lstm_layer_dim = 0
        self.config.print("Loading BiRNN: %s" % self, level=4)

    def __str__(self):
        return "%s lstm_layers: %d, lstm_layer_dim: %d, embedding_layers: %d, embedding_layer_dim: %d, " \
               "max_length: %d, rnn: %s, activation: %s, init: %s, dropout: %f, indexed_dim: %s, indexed_num: %s, "\
               "params: %s" % (
                "/".join(self.save_path), self.lstm_layers, self.lstm_layer_dim, self.embedding_layers,
                self.embedding_layer_dim, self.max_length, self.rnn_builder, self.activation, self.init, self.dropout,
                self.indexed_dim, self.indexed_num, list(self.params.keys()))

    def sub_models(self):
        return [self.mlp]


class EmptyRNN(BiRNN):
    def init_params(self, indexed_dim, indexed_num):
        pass

    def init_features(self, embeddings, train=False):
        pass

    def evaluate(self, indices):
        return []

    def save_sub_model(self, d, *args):
        return []

    def load_sub_model(self, d, *args, **kwargs):
        super().load_sub_model(d, *args)
        self.args.lstm_layers = self.lstm_layers = self.args.lstm_layer_dim = self.lstm_layer_dim = 0


class HighwayRNN(BiRNN):
    def init_rnn_params(self, indexed_dim):
        params = []
        for i in range(self.lstm_layers):
            for n in "f", "b":
                input_dim = self.lstm_layer_dim if i or self.embedding_layers else indexed_dim
                rnn = self.rnn_builder()(1, input_dim, self.lstm_layer_dim, self.model)
                self.params["rnn%d%s" % (i, n)] = rnn
                params += [p for l in rnn.get_parameters() for p in l]
                for p, dim in (("Wr", (self.lstm_layer_dim, input_dim + self.lstm_layer_dim)),
                               ("br", self.lstm_layer_dim), ("Wh", (self.lstm_layer_dim, input_dim))):
                    param = self.model.add_parameters(dim, init=self.init()())
                    self.params["%s%d%s" % (p, i, n)] = param
                    params.append(param)
        return params

    def transduce(self, inputs, train):
        xs = inputs[:self.max_length]
        if not xs:
            return []
        for i in range(self.lstm_layers):
            for n, d in ("f", 1), ("b", -1):
                Wr, br, Wh = [self.params["%s%d%s" % (p, i, n)] for p in ("Wr", "br", "Wh")]
                hs_ = self.params["rnn%d%s" % (i, n)].initial_state().transduce(xs[::d])
                hs = [hs_[0]]
                for t in range(1, len(hs_)):
                    r = dy.logistic(Wr * dy.concatenate([hs[t - 1], xs[t]]) + br)
                    hs.append(dy.cmult(r, hs_[t]) + dy.cmult(1 - r, Wh * xs[t]))
                xs = hs
                if train:
                    x = dy.dropout_dim(dy.concatenate(xs, 1), 1, self.dropout)
                    xs = [dy.pick(x, i, 1) for i in range(len(xs))]
        return xs


class HierarchicalBiRNN(BiRNN):
    RNN_NAMES = "lrnn", "rrnn"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.internal_reps = {}

    def add_node(self, i):
        self.internal_reps[i] = [self.params[n].initial_state() for n in self.RNN_NAMES]

    def add_edge(self, i, j, direction):
        self.internal_reps[i][direction] = self.internal_reps[i][direction].add_input(self.get_representation(j))

    def transition(self, action):
        if self.params:
            if action.node:
                self.add_node(action.node.index)
            if action.edge:
                self.add_edge(action.edge.parent.index, action.edge.child.index, "RIGHT" in action.type)

    def init_features(self, embeddings, train=False):
        super().init_features(embeddings, train)
        self.internal_reps.clear()
        if self.params:
            self.add_node(0)  # Root

    def init_rnn_params(self, indexed_dim):
        params = super().init_rnn_params(indexed_dim)
        for name in self.RNN_NAMES:
            rnn = self.rnn_builder()(1, self.lstm_layer_dim, self.lstm_layer_dim / 2, self.model)
            self.params[name] = rnn
            params += [p for l in rnn.get_parameters() for p in l]
        return params

    def get_representation(self, i):
        states = self.internal_reps.get(i)
        if states:
            return dy.concatenate([s.output() or self.empty_rep[:self.lstm_layer_dim / 2] for s in states])
        return super().get_representation(i - 1)  # Terminal index is (node index-1)



import numpy as np
from configargparse import ArgParser

from tupa.model import Model


def load_model(filename):
    model = Model(filename=filename)
    model.load()
    return model


def save_model(model, filename):
    out_file = filename + ".npz"
    np.savez_compressed(out_file, **model.all_params())
    print("Wrote '%s'" % out_file)


def main():
    argparser = ArgParser(description="Load TUPA model and export as .npz file.")
    argparser.add_argument("models", nargs="+", help="model file basename(s) to load")
    args = argparser.parse_args()
    for filename in args.models:
        model = load_model(filename)
        save_model(model, filename)
        model.config.save(filename)


if __name__ == "__main__":
    main()



#!/usr/bin/env python3

import argparse
import csv
from operator import attrgetter

from tqdm import tqdm
from ucca import textutil
from spacy.symbols import NAMES

desc = """Create vocabulary file from spaCy model."""


def main(args):
    if args.model_name:
        textutil.models[args.lang] = args.model_name
    vocab = list(enumerate(NAMES)) + \
        [(l.orth, l.orth_) for l in sorted(textutil.get_vocab(lang=args.lang), key=attrgetter("prob"), reverse=True)]
    out_file = args.out_file or textutil.models[args.lang] + ".csv"
    with open(out_file, "w", encoding="utf-8") as f:
        csv.writer(f).writerows(tqdm(vocab, unit=" words", desc="Writing " + out_file))


if __name__ == '__main__':
    argparser = argparse.ArgumentParser(description=desc)
    argparser.add_argument("model_name", nargs="?", help="name of spaCy model")
    argparser.add_argument("out_file", nargs="?", help="filename to create")
    argparser.add_argument("-l", "--lang", default="en", help="small two-letter language code to use for NLP model")
    main(argparser.parse_args())



from .feature_extractor import FeatureExtractor

FEATURE_TEMPLATES = (
    # unigrams (Zhang and Clark 2009):
    "s0tde", "s0we", "s0#e", "s0^e", "s0$e", "s1tde", "s1we", "s1#e", "s1^e", "s1$e",
    "s2tde", "s2we", "s2#e", "s2^e", "s2$e", "s3tde", "s3we", "s3#e", "s3^e", "s3$e",
    "b0wtd", "b1wtd", "b2wtd", "b3wtd", "s0lwe", "s0rwe", "s1lwe", "s1rwe",
    # bigrams (Zhang and Clark 2009):
    "s0ws1w", "s0ws1e", "s0es1w", "s0es1e", "s0wb0w", "s0wb0td",
    "s0eb0w", "s0eb0td", "s1wb0w", "s1wb0td", "s1eb0w", "s1eb0td", "b0wb1w", "b0wb1td", "b0tdb1w", "b0tdb1td",
    # trigrams (Zhang and Clark 2009):
    "s0es1es2#", "s0es1es2e", "s0es1es2e", "s0es1eb0#", "s0es1eb0td",
    "s0es1#b0#", "s0es1#b0td", "s0#s1es2e", "s0#s1eb0td",
    # extended (Zhu et al. 2013):
    "s0llwe", "s0lrwe", "s0rlwe", "s0rrwe", "s1llwe", "s1lrwe", "s1rlwe", "s1rrwe",
    # parents:
    "s0Lwe", "s0Rwe", "s1Lwe", "s1Rwe", "b0Lwe", "b0Rwe",
    # separator (Zhu et al. 2013):
    "s0wp", "s0wep", "s0wq", "s0weq", "s0es1ep", "s0es1eq", "s1wp", "s1wep", "s1wq", "s1weq",
    # disco, unigrams (Maier 2015):
    "s0xwe", "s1xwe", "s2xwe", "s3xwe", "s0xtde", "s1xtde", "s2xtde", "s3xtde", "s0xy", "s1xy", "s2xy", "s3xy",
    # disco, bigrams (Maier 2015):
    "s0xs1e", "s0xs1w", "s0xs1x", "s0ws1x", "s0es1x", "s0xs2e", "s0xs2w", "s0xs2x", "s0es2x",
    "s0ys1y", "s0ys2y", "s0xb0td", "s0xb0w",
    # counts (Tokgöz and Eryiğit 2015):
    "s0P", "s0C", "s0wP", "s0wC", "b0P", "b0C", "b0wP", "b0wC",
    # existing edges (Tokgöz and Eryiğit 2015):
    "s0s1x", "s1s0x", "s0b0x", "b0s0x", "s0b0e", "b0s0e",
    # dependency distance:
    "s0s1d", "s0b0d",
    # past actions (Tokgöz and Eryiğit 2015):
    "a0Ae", "a1Ae",
    # implicit and remote
    "s0I", "s0E", "s0M", "s0wI", "s0wE", "s0wM", "b0I", "b0E", "b0M", "b0wI", "b0wE", "b0wM",
    # node labels
    "s0n", "s0wn", "s0c", "b0n", "b0wn", "b0c", "s0Ln", "s0Rn", "s1Ln", "s1Rn", "b0Ln", "b0Rn",
    # height
    "s0h", "s1h", "b0h", "b1h",
    # NER
    "s0NT", "s1NT", "b0NT", "b1NT",
)


class SparseFeatureExtractor(FeatureExtractor):
    """
    Object to extract features from the parser state to be used in action classification
    To be used with SparsePerceptron classifier.
    """

    def __init__(self, omit_features=None):
        super().__init__(feature_templates=FEATURE_TEMPLATES, omit_features=omit_features)

    def extract_features(self, state):
        """
        Calculate feature values according to current state
        :param state: current state of the parser
        :return dict of feature name -> value
        """
        features = {
            "b": 1,  # Bias
            "n/t": state.node_ratio(),  # number of nodes divided by number of terminals
        }
        for feature_template in self.feature_templates:
            values = feature_template.extract(state)
            if values:
                features["%s=%s" % (feature_template.name, " ".join(map(str, values)))] = 1
        return features



"""Testing code for the tupa.config module, unit-testing only."""

import pytest

from tupa.config import Config


@pytest.fixture
def config():
    c = Config("", "-m", "test")
    c.update({"no_node_labels": True, "evaluate": True, "minibatch_size": 50})
    c.update_hyperparams(shared={"layer_dim": 50})
    return c


def test_params(config):
    d = {"max_words_external": 100, "word_dim_external": 100, "optimizer": "sgd", "layer_dim": 100, "layers": 1,
         "lstm_layer_dim": 100, "lstm_layers": 1}
    config.update(d)
    for attr, value in d.items():
        assert getattr(config.args, attr) == value, attr


def test_hyperparams(config):
    assert config.hyperparams.shared.layer_dim == 50, "--hyperparams=shared=--layer-dim=50"
    d = {"max_words_external": 100, "word_dim_external": 100, "optimizer": "sgd", "layer_dim": 100, "layers": 1}
    config.update(d)
    config.update_hyperparams(shared={"lstm_layer_dim": 100, "lstm_layers": 1}, ucca={"word_dim": 300},
                              **{"ucca.en": {"dep_dim": 1}})
    assert config.hyperparams.shared.lstm_layer_dim == 100, "shared --lstm-layer-dim=100"
    assert config.hyperparams.shared.lstm_layers == 1, "shared --lstm-layers=1"
    assert config.hyperparams.shared.minibatch_size == 50, "--minibatch-size=50"
    assert config.hyperparams.specific["ucca"].word_dim == 300, "ucca --word-dim=300"
    assert config.hyperparams.specific["ucca.en"].dep_dim == config.hyperparams.specific["ucca"]["en"].dep_dim == 1, \
        "ucca.en --dep-dim=1"
    assert config.hyperparams.specific["ucca"].minibatch_size == 50, "--minibatch-size=50"
    for attr, value in d.items():
        assert getattr(config.hyperparams.shared, attr) == value, attr


def test_boolean_params(config):
    assert config.args.evaluate
    assert not config.args.verify



from collections import OrderedDict


class SubModel:
    def __init__(self, params=None, save_path=(), shared=False, copy_shared=False):
        self.params = OrderedDict() if params is None else params  # string (param identifier) -> parameter
        self.save_path = save_path
        self.shared = shared
        self.copy_shared = copy_shared

    def save_sub_model(self, d, *args):
        self.get_sub_dict(d).update(args + (("param_keys", list(self.params.keys())),))
        return list(self.params.values())

    def load_sub_model(self, d, *args, load_path=None):
        d = self.get_sub_dict(d, load_path=load_path)
        param_keys = d.get("param_keys", ())
        assert len(param_keys) <= len(args), "%s loaded values: expected %d, got %d" % ("/".join(self.save_path),
                                                                                        len(param_keys), len(args))
        self.params.clear()
        self.params.update(zip(param_keys, args))
        return d

    def get_sub_dict(self, d, load_path=None):
        for element in load_path or self.save_path:
            d = d.setdefault(element, OrderedDict())
        return d

    def params_str(self):
        return "/".join(self.save_path) + (": " if self.save_path else "") + ", ".join(self.params.keys())

    def invalidate_caches(self):
        for model in self.sub_models():
            model.invalidate_caches()

    def sub_models(self):
        return ()



import os
import shlex
from copy import deepcopy

import dynet_config
import numpy as np
from configargparse import ArgParser, Namespace, ArgumentDefaultsHelpFormatter, SUPPRESS
from logbook import Logger, FileHandler, StderrHandler
from semstr.cfgutil import Singleton, add_verbose_arg, add_boolean_option, get_group_arg_names
from semstr.convert import UCCA_EXT, CONVERTERS
from ucca import constructions

from tupa.classifiers.nn.constants import *
from tupa.model_util import load_enum

# Classifiers

SPARSE = "sparse"
MLP = "mlp"
BIRNN = "bilstm"
HIGHWAY_RNN = "highway"
HIERARCHICAL_RNN = "hbirnn"
NOOP = "noop"
NN_CLASSIFIERS = (MLP, BIRNN, HIGHWAY_RNN, HIERARCHICAL_RNN)
CLASSIFIERS = (SPARSE, MLP, BIRNN, HIGHWAY_RNN, HIERARCHICAL_RNN, NOOP)

FEATURE_PROPERTIES = "wmtudhencpqxyAPCIEMNT#^$"

# Swap types
REGULAR = "regular"
COMPOUND = "compound"

# Input/output formats
FORMATS = ["ucca"] + list(CONVERTERS)
FILE_FORMATS = [e.lstrip(".") for e in UCCA_EXT] + FORMATS

# Required number of edge labels per format
EDGE_LABELS_NUM = {"amr": 110, "sdp": 70, "conllu": 60}
SPARSE_ARG_NAMES = set()
NN_ARG_NAMES = set()
DYNET_ARG_NAMES = set()
RESTORED_ARGS = set()

SEPARATOR = "."


def add_param_arguments(ap=None, arg_default=None):  # arguments with possible format-specific parameter values

    def add_argument(a, *args, **kwargs):
        return a.add_argument(*args, **kwargs)

    def add(a, *args, default=None, func=add_argument, **kwargs):
        arg = func(a, *args, default=default if arg_default is None else arg_default, **kwargs)
        try:
            RESTORED_ARGS.add(arg.dest)
        except AttributeError:
            RESTORED_ARGS.update(get_group_arg_names(arg))

    def add_boolean(a, *args, **kwargs):
        add(a, *args, func=add_boolean_option, **kwargs)

    if not ap:
        ap = ArgParser()

    group = ap.add_argument_group(title="Node labels")
    add(group, "--max-node-labels", type=int, default=0, help="max number of node labels to allow")
    add(group, "--max-node-categories", type=int, default=0, help="max node categories to allow")
    add(group, "--min-node-label-count", type=int, default=2, help="min number of occurrences for a label")
    add_boolean(group, "use-gold-node-labels", "gold node labels when parsing")
    add_boolean(group, "wikification", "use Spotlight to wikify any named node")
    add_boolean(group, "node-labels", "prediction of node labels, if supported by format", default=True)

    group = ap.add_argument_group(title="Structural constraints")
    add_boolean(group, "linkage", "linkage nodes and edges")
    add_boolean(group, "implicit", "implicit nodes and edges")
    add_boolean(group, "remote", "remote edges", default=True)
    add_boolean(group, "constraints", "scheme-specific rules", default=True)
    add_boolean(group, "require-connected", "constraint that output graph must be connected")
    add(group, "--orphan-label", default="orphan", help="edge label to use for nodes without parents")
    add(group, "--max-action-ratio", type=float, default=100, help="max action/terminal ratio")
    add(group, "--max-node-ratio", type=float, default=10, help="max node/terminal ratio")
    add(group, "--max-height", type=int, default=20, help="max graph height")

    group = ap.add_mutually_exclusive_group()
    add(group, "--swap", choices=(REGULAR, COMPOUND), default=REGULAR, help="swap transitions")
    add(group, "--no-swap", action="store_false", dest="swap", help="exclude swap transitions")
    add(ap, "--max-swap", type=int, default=15, help="if compound swap enabled, maximum swap size")

    group = ap.add_argument_group(title="General classifier training parameters")
    add(group, "--learning-rate", type=float, help="rate for model weight updates (default: by trainer/1)")
    add(group, "--learning-rate-decay", type=float, default=0, help="learning rate decay per iteration")
    add(group, "--swap-importance", type=float, default=1, help="learning rate factor for Swap")
    add(group, "--max-training-per-format", type=int, help="max number of training passages per format per iteration")
    add_boolean(group, "missing-node-features", "allow node features to be missing if not available", default=True)
    add(group, "--omit-features", help="string of feature properties to omit, out of " + FEATURE_PROPERTIES)
    add_boolean(group, "curriculum", "sort training passages by action prediction accuracy in previous epoch")

    group = ap.add_argument_group(title="Perceptron parameters")
    add(group, "--min-update", type=int, default=5, help="minimum #updates for using a feature")
    SPARSE_ARG_NAMES.update(get_group_arg_names(group))

    group = ap.add_argument_group(title="Neural network parameters")
    add(group, "--word-dim-external", type=int, default=300, help="dimension for external word embeddings")
    add(group, "--word-vectors", help="file to load external word embeddings from (default: GloVe)")
    add(group, "--vocab", help="file mapping integer ID to word form (to avoid loading spaCy), or '-' to use word form")
    add_boolean(group, "update-word-vectors", "external word vectors in training parameters", default=True)
    add(group, "--word-dim", type=int, default=0, help="dimension for learned word embeddings")
    add(group, "--lemma-dim", type=int, default=200, help="dimension for lemma embeddings")
    add(group, "--tag-dim", type=int, default=20, help="dimension for fine POS tag embeddings")
    add(group, "--pos-dim", type=int, default=20, help="dimension for coarse/universal POS tag embeddings")
    add(group, "--dep-dim", type=int, default=10, help="dimension for dependency relation embeddings")
    add(group, "--edge-label-dim", type=int, default=20, help="dimension for edge label embeddings")
    add(group, "--node-label-dim", type=int, default=0, help="dimension for node label embeddings")
    add(group, "--node-category-dim", type=int, default=0, help="dimension for node category embeddings")
    add(group, "--punct-dim", type=int, default=1, help="dimension for separator punctuation embeddings")
    add(group, "--action-dim", type=int, default=3, help="dimension for input action type embeddings")
    add(group, "--ner-dim", type=int, default=3, help="dimension for input entity type embeddings")
    add(group, "--shape-dim", type=int, default=3, help="dimension for word shape embeddings")
    add(group, "--prefix-dim", type=int, default=2, help="dimension for word prefix embeddings")
    add(group, "--suffix-dim", type=int, default=3, help="dimension for word suffix embeddings")
    add(group, "--output-dim", type=int, default=50, help="dimension for output action embeddings")
    add(group, "--layer-dim", type=int, default=50, help="dimension for hidden layers")
    add(group, "--layers", type=int, default=2, help="number of hidden layers")
    add(group, "--lstm-layer-dim", type=int, default=500, help="dimension for LSTM hidden layers")
    add(group, "--lstm-layers", type=int, default=0, help="number of LSTM hidden layers")
    add(group, "--embedding-layer-dim", type=int, default=500, help="dimension for layers before LSTM")
    add(group, "--embedding-layers", type=int, default=1, help="number of layers before LSTM")
    add(group, "--activation", choices=ACTIVATIONS, default=DEFAULT_ACTIVATION, help="activation function")
    add(group, "--init", choices=INITIALIZERS, default=DEFAULT_INITIALIZER, help="weight initialization")
    add(group, "--minibatch-size", type=int, default=200, help="mini-batch size for optimization")
    add(group, "--optimizer", choices=TRAINERS, default=DEFAULT_TRAINER, help="algorithm for optimization")
    add(group, "--loss", choices=LOSSES, default=DEFAULT_LOSS, help="loss function for training")
    add(group, "--max-words-external", type=int, default=250000, help="max external word vectors to use")
    add(group, "--max-words", type=int, default=10000, help="max number of words to keep embeddings for")
    add(group, "--max-lemmas", type=int, default=3000, help="max number of lemmas to keep embeddings for")
    add(group, "--max-tags", type=int, default=100, help="max number of fine POS tags to keep embeddings for")
    add(group, "--max-pos", type=int, default=100, help="max number of coarse POS tags to keep embeddings for")
    add(group, "--max-deps", type=int, default=100, help="max number of dep labels to keep embeddings for")
    add(group, "--max-edge-labels", type=int, default=15, help="max number of edge labels for embeddings")
    add(group, "--max-puncts", type=int, default=5, help="max number of punctuations for embeddings")
    add(group, "--max-action-types", type=int, default=10, help="max number of action types for embeddings")
    add(group, "--max-action-labels", type=int, default=100, help="max number of action labels to allow")
    add(group, "--max-ner-types", type=int, default=18, help="max number of entity types to allow")
    add(group, "--max-shapes", type=int, default=30, help="max number of word shapes to allow")
    add(group, "--max-prefixes", type=int, default=30, help="max number of 1-character word prefixes to allow")
    add(group, "--max-suffixes", type=int, default=500, help="max number of 3-character word suffixes to allow")
    add(group, "--word-dropout", type=float, default=0.2, help="word dropout parameter")
    add(group, "--word-dropout-external", type=float, default=0, help="word dropout for word vectors")
    add(group, "--lemma-dropout", type=float, default=0.2, help="lemma dropout parameter")
    add(group, "--tag-dropout", type=float, default=0.2, help="fine POS tag dropout parameter")
    add(group, "--pos-dropout", type=float, default=0.2, help="coarse POS tag dropout parameter")
    add(group, "--dep-dropout", type=float, default=0.5, help="dependency label dropout parameter")
    add(group, "--node-label-dropout", type=float, default=0.2, help="node label dropout parameter")
    add(group, "--node-dropout", type=float, default=0.1, help="probability to drop features for a whole node")
    add(group, "--dropout", type=float, default=0.4, help="dropout parameter between layers")
    add(group, "--max-length", type=int, default=120, help="maximum length of input sentence")
    add(group, "--rnn", choices=["None"] + list(RNNS), default=DEFAULT_RNN, help="type of recurrent neural network")
    add(group, "--gated", type=int, nargs="?", default=2, help="gated input to BiRNN and MLP")
    NN_ARG_NAMES.update(get_group_arg_names(group))
    return ap


class FallbackNamespace(Namespace):
    def __init__(self, fallback, kwargs=None):
        super().__init__(**(kwargs or {}))
        self._fallback = fallback
        self._children = {}

    def __getattr__(self, item):
        if item.startswith("_"):
            return getattr(super(), item)
        return getattr(super(), item, getattr(self._fallback, item))

    def __getitem__(self, item):
        if item:
            name, _, rest = item.partition(SEPARATOR)
            return self._children.setdefault(name, FallbackNamespace(self))[rest]
        return self

    def vars(self):
        return {k: v for k, v in vars(self).items() if not k.startswith("_")}

    def items(self):
        return self.vars().items()

    def update(self, kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def traverse(self, prefix=None):
        if prefix and self.vars():
            yield (prefix, self)
        for name, child in self._children.items():
            yield from child.traverse(SEPARATOR.join(filter(None, (prefix, name))))


class Hyperparams:
    def __init__(self, parent, shared=None, **kwargs):
        self.shared = FallbackNamespace(parent, shared)
        self.specific = FallbackNamespace(parent)
        for name, args in kwargs.items():
            self.specific[name].update(args)

    def items(self):
        return ([("shared", self.shared)] if self.shared.vars() else []) + list(self.specific.traverse())


class HyperparamsInitializer:
    def __init__(self, name=None, *args, **kwargs):
        """
        :param name: name of hyperparams subset
        :param args: raw arg strings
        :param kwargs: parsed and initialized values
        """
        self.name = name
        self.str_args = list(args) + ["--%s %s" % (k.replace("_", "-"), v) for k, v in kwargs.items()]
        self.args = vars(add_param_arguments(arg_default=SUPPRESS).parse_args(args))
        self.args.update(kwargs)

    def __str__(self):
        return '"%s"' % " ".join([self.name] + list(self.str_args))

    def __bool__(self):
        return bool(self.str_args)
    
    @classmethod
    def action(cls, args):
        return cls(*args.replace("=", " ").split())


class Iterations:
    def __init__(self, args):
        try:
            epochs, *hyperparams = args.replace("=", " ").split()
        except (AttributeError, ValueError):
            epochs, *hyperparams = args,
        self.epochs, self.hyperparams = int(epochs), HyperparamsInitializer(str(epochs), *hyperparams)

    def __str__(self):
        return str(self.hyperparams or self.epochs)


class Config(object, metaclass=Singleton):
    def __init__(self, *args):
        self.arg_parser = ap = ArgParser(description="Transition-based parser for UCCA.",
                                         formatter_class=ArgumentDefaultsHelpFormatter)

        add_boolean_option(ap, "use-bert", default=False, description="whether to use bert embeddings")
        ap.add_argument("--bert-model", choices=["bert-base-uncased", "bert-large-uncased", "bert-base-cased",
                                                 "bert-large-cased", "bert-base-multilingual-cased"],
                        default="bert-base-multilingual-cased")
        ap.add_argument("--bert-layers", type=int, nargs='+', default=[-1, -2, -3, -4])
        ap.add_argument("--bert-layers-pooling", choices=["weighted", "sum", "concat"], default="weighted")
        ap.add_argument("--bert-token-align-by", choices=["first", "sum", "mean"], default="sum")
        ap.add_argument("--bert-multilingual", choices=[0], type=int)
        add_boolean_option(ap, "bert-use-default-word-embeddings", default=False,
                           description="whether to use default word embeddings")
        ap.add_argument("--bert-dropout", type=float, default=0, choices=np.linspace(0, 0.9, num=10))

        ap.add_argument("passages", nargs="*", help="passage files/directories to test on/parse")
        ap.add_argument("--version", action="version", version="")
        ap.add_argument("-C", "--config", is_config_file=True, help="configuration file to get arguments from")
        ap.add_argument("-m", "--models", nargs="+", help="model file basename(s) to load/save, ensemble if >1 "
                                                          "(default: <format>_<model_type>")
        ap.add_argument("-c", "--classifier", choices=CLASSIFIERS, default=BIRNN, help="model type")
        ap.add_argument("-B", "--beam", type=int, choices=(1,), default=1, help="beam size for beam search")
        add_boolean_option(ap, "evaluate", "evaluation of parsed passages", short="e")
        add_verbose_arg(ap, help="detailed parse output")
        constructions.add_argument(ap)
        add_boolean_option(ap, "sentences", "split to sentences")
        add_boolean_option(ap, "paragraphs", "split to paragraphs")
        ap.add_argument("--timeout", type=float, help="max number of seconds to wait for a single passage")

        group = ap.add_argument_group(title="Training parameters")
        group.add_argument("-t", "--train", nargs="+", default=(), help="passage files/directories to train on")
        group.add_argument("-d", "--dev", nargs="+", default=(), help="passage files/directories to tune on")
        group.add_argument("-I", "--iterations", nargs="+", type=Iterations,
                           default=(Iterations(50), Iterations("100 --optimizer=" + EXTRA_TRAINER)),
                           help="number of training iterations along with optional hyperparameters per part")
        group.add_argument("--folds", type=int, choices=(3, 5, 10), help="#folds for cross validation")
        group.add_argument("--seed", type=int, default=1, help="random number generator seed")
        add_boolean_option(group, "early-update", "early update procedure (finish example on first error)")
        group.add_argument("--save-every", type=int, help="every this many passages, evaluate on dev and save model")
        add_boolean_option(group, "eval-test", "evaluate on test whenever evaluating on dev, but keep results hidden")
        add_boolean_option(group, "ignore-case", "pre-convert all input files to lower-case in training and test")

        group = ap.add_argument_group(title="Output files")
        group.add_argument("-o", "--outdir", default=".", help="output directory for parsed files")
        group.add_argument("-p", "--prefix", default="", help="output filename prefix")
        add_boolean_option(group, "write", "writing parsed output to files", default=True, short_no="W")
        group.add_argument("-j", "--join", help="if output format is textual, write all to one file with this basename")
        group.add_argument("-l", "--log", help="output log file (default: model filename + .log)")
        group.add_argument("--devscores", help="output CSV file for dev scores (default: model filename + .dev.csv)")
        group.add_argument("--testscores", help="output CSV file for test scores (default: model filename + .test.csv)")
        group.add_argument("--action-stats", help="output CSV file for action statistics")
        add_boolean_option(group, "normalize", "apply normalizations to output in case format is UCCA", default=False)
        ap.add_argument("-f", "--formats", nargs="+", choices=FILE_FORMATS, default=(),
                        help="input formats for creating all parameters before training starts "
                             "(otherwise created dynamically based on filename suffix), "
                             "and output formats for written files (each will be written; default: UCCA XML)")
        ap.add_argument("-u", "--unlabeled", nargs="*", choices=FORMATS, help="to ignore labels in")
        ap.add_argument("--lang", default="en", help="two-letter language code to use as the default language")
        add_boolean_option(ap, "multilingual", "separate model parameters per language (passage.attrib['lang'])")

        group = ap.add_argument_group(title="Sanity checks")
        add_boolean_option(group, "check-loops", "check for parser state loop")
        add_boolean_option(group, "verify", "check for oracle reproducing original passage")
        add_boolean_option(group, "validate-oracle", "require oracle output to respect constraints", default=True)
        add_param_arguments(ap)

        group = ap.add_argument_group(title="DyNet parameters")
        group.add_argument("--dynet-mem", help="memory for dynet")
        group.add_argument("--dynet-weight-decay", type=float, default=1e-5, help="weight decay for parameters")
        add_boolean_option(group, "dynet-apply-weight-decay-on-load", "workaround for clab/dynet#1206", default=False)
        add_boolean_option(group, "dynet-gpu", "GPU for training")
        group.add_argument("--dynet-gpus", type=int, default=1, help="how many GPUs you want to use")
        add_boolean_option(group, "dynet-autobatch", "auto-batching of training examples")
        DYNET_ARG_NAMES.update(get_group_arg_names(group))

        ap.add_argument("-H", "--hyperparams", type=HyperparamsInitializer.action, nargs="*",
                        help="shared hyperparameters or hyperparameters for specific formats, "
                             'e.g., "shared --lstm-layer-dim=100 --lstm-layers=1" "ucca --word-dim=300"',
                        default=[HyperparamsInitializer.action("shared --lstm-layers 2")])
        ap.add_argument("--copy-shared", nargs="*", choices=FORMATS, help="formats whose parameters shall be "
                                                                          "copied from loaded shared parameters")
        self.args = FallbackNamespace(ap.parse_args(args if args else None))

        if self.args.config:
            print("Loading configuration from '%s'." % self.args.config)

        if self.args.passages and self.args.write:
            os.makedirs(self.args.outdir, exist_ok=True)

        if self.args.models:
            if not self.args.log:
                self.args.log = self.args.models[0] + ".log"
            if self.args.dev and not self.args.devscores:
                self.args.devscores = self.args.models[0] + ".dev.csv"
            if self.args.passages and not self.args.testscores:
                self.args.testscores = self.args.models[0] + ".test.csv"
        elif not self.args.log:
            self.args.log = "parse.log"
        self.sub_configs = []  # Copies to be stored in Models so that they do not interfere with each other
        self._logger = self.format = self.hyperparams = self.iteration_hyperparams = None
        self._vocab = {}
        self.original_values = {}
        self.random = np.random
        self.update()

    def create_original_values(self, args=None):
        return {attr: getattr(self.args, attr) if args is None else args[attr]
                for attr in RESTORED_ARGS if args is None or attr in args}

    def set_format(self, f=None, update=False, recursive=True):
        if f in (None, "text") and not self.format:  # In update or parsing UCCA (with no extra["format"]) or plain text
            f = "ucca"  # Default output format is UCCA
        if update or self.format != f:
            if f not in (None, "text"):
                self.format = f
            self.update_by_hyperparams()
        if recursive:
            for config in self.descendants():
                config.set_format(f=f, update=update, recursive=False)

    def descendants(self):
        ret = []
        configs = [self]
        while configs:
            c = configs.pop(0)
            ret += c.sub_configs
            configs += c.sub_configs
        return ret

    def is_unlabeled(self, f=None):
        # If just -u or --unlabeled is given then its value is [], and we want to treat that as "all formats"
        # If not given at all it is None, and we want to treat that as "no format"
        return self.args.unlabeled == [] or (f or self.format) in (self.args.unlabeled or ())

    def max_actions_unlabeled(self):
        return 6 + (  # Shift Node Reduce LeftEdge RightEdge Finish
            3 if self.args.remote else 0) + (  # RemoteNode LeftRemote RightRemote
            1 if self.args.swap == REGULAR else (self.args.max_swap if self.args.swap == COMPOUND else 0)) + (  # Swap
            1 if self.args.implicit else 0) + (  # Implicit
            2 if self.args.node_labels and not self.args.use_gold_node_labels else 0)  # Label x 2

    def set_dynet_arguments(self):
        self.random.seed(self.args.seed)
        kwargs = dict(random_seed=self.args.seed)
        if self.args.dynet_mem:
            kwargs.update(mem=self.args.dynet_mem)
        if self.args.dynet_weight_decay:
            kwargs.update(weight_decay=self.args.dynet_weight_decay)
        if self.args.dynet_gpus and self.args.dynet_gpus != 1:
            kwargs.update(requested_gpus=self.args.dynet_gpus)
        if self.args.dynet_autobatch:
            kwargs.update(autobatch=True)
        dynet_config.set(**kwargs)
        if self.args.dynet_gpu:
            dynet_config.set_gpu()

    def update(self, params=None):
        if params:
            for name, value in params.items():
                setattr(self.args, name, value)
        self.original_values.update(self.create_original_values(params))
        self.hyperparams = self.create_hyperparams()
        for f, num in EDGE_LABELS_NUM.items():
            self.hyperparams.specific[f].max_edge_labels = num
        amr_hyperparams = self.hyperparams.specific["amr"]
        for k, v in dict(node_label_dim=20, max_node_labels=1000, node_category_dim=5, max_node_categories=25).items():
            if k not in amr_hyperparams and not getattr(amr_hyperparams, k, None):
                setattr(amr_hyperparams, k, v)
        self.set_format(update=True)
        self.set_dynet_arguments()

    def create_hyperparams(self):
        return Hyperparams(parent=self.args, **{h.name: h.args for h in self.args.hyperparams or ()})

    def update_hyperparams(self, **kwargs):
        self.update(dict(hyperparams=[HyperparamsInitializer(k, **v) for k, v in kwargs.items()]))

    def update_iteration(self, iteration, print_message=True, recursive=True):
        if iteration.hyperparams:
            if print_message:
                print("Updating: %s" % iteration.hyperparams)
            self.iteration_hyperparams = iteration.hyperparams.args
            self.update_by_hyperparams()
            if recursive:
                for config in self.descendants():
                    config.update_iteration(iteration, print_message=False, recursive=False)

    def update_by_hyperparams(self):
        format_values = dict(self.original_values)
        for hyperparams in (self.iteration_hyperparams, self.hyperparams.specific[self.format]):
            if hyperparams:
                format_values.update({k: v for k, v in hyperparams.items() if not k.startswith("_")})
        for attr, value in sorted(format_values.items()):
            self.print("Setting %s=%s" % (attr, value))
            setattr(self.args, attr, value)
        if self.format != "amr":
            self.args.node_labels = False
            self.args.node_label_dim = self.args.max_node_labels = \
                self.args.node_category_dim = self.args.max_node_categories = 0
        if self.is_unlabeled():
            self.args.max_edge_labels = self.args.edge_label_dim = 0
            self.args.max_action_labels = self.max_actions_unlabeled()
        else:
            self.args.max_action_labels = max(self.args.max_action_labels, 6 * self.args.max_edge_labels)

    @property
    def line_end(self):
        return "\n" if self.args.verbose > 2 else " "  # show all in one line unless verbose

    @property
    def passage_word(self):
        return "sentence" if self.args.sentences else "paragraph" if self.args.paragraphs else "passage"

    @property
    def passages_word(self):
        return " %ss" % self.passage_word

    def log(self, message):
        try:
            if self._logger is None:
                FileHandler(self.args.log,
                            format_string="{record.time:%Y-%m-%d %H:%M:%S} {record.message}").push_application()
                if self.args.verbose > 1:
                    StderrHandler(bubble=True).push_application()
                self._logger = Logger("tupa")
            self._logger.warn(message)
        except OSError:
            pass

    def vocab(self, filename=None, lang=None):
        if filename is None:
            args = self.hyperparams.specific[lang] if lang else self.args
            filename = args.vocab
        if not filename:
            return None
        vocab = self._vocab.get(filename)
        if vocab:
            return vocab
        vocab = load_enum(filename)
        self._vocab[filename] = vocab
        return vocab

    def print(self, message, level=3):
        if self.args.verbose >= level:
            try:
                print(message() if hasattr(message, "__call__") else message, flush=True)
            except UnicodeEncodeError:
                pass

    def save(self, filename):
        out_file = filename + ".yml"
        print("Saving configuration to '%s'." % out_file)
        with open(out_file, "w") as f:
            name = None
            values = []
            for arg in shlex.split(str(self), "--") + ["--"]:
                if arg.startswith("--"):
                    if name and name not in ("train", "dev"):
                        if len(values) > 1:
                            values[0] = "[" + values[0]
                            values[-1] += "]"
                        elif name.startswith("no-"):
                            name = name[3:]
                            values = ["false"]
                        print("%s: %s" % (name, ", ".join(values) or "true"), file=f)
                    name = arg[2:]
                    values = []
                else:
                    values.append(arg)

    def copy(self):
        cls = self.__class__
        ret = cls.__new__(cls)
        ret.arg_parser = self.arg_parser
        ret.args = deepcopy(self.args)
        ret.original_values = deepcopy(self.original_values)
        ret.hyperparams = deepcopy(self.hyperparams)
        ret.iteration_hyperparams = deepcopy(self.iteration_hyperparams)
        ret.format = self.format
        ret.random = self.random
        ret._logger = self._logger
        ret._vocab = dict(self._vocab)
        ret.sub_configs = []
        self.sub_configs.append(ret)
        return ret

    def args_str(self, args):
        return ["--" + ("no-" if v is False else "") + k.replace("_", "-") +
                ("" if v is False or v is True else
                 (" " + str(" ".join(map(str, v)) if hasattr(v, "__iter__") and not isinstance(v, str) else v)))
                for (k, v) in sorted(args.items()) if
                v not in (None, (), "", self.arg_parser.get_default(k))
                and not k.startswith("_")
                and (args.node_labels or ("node_label" not in k and "node_categor" not in k))
                and (args.swap or "swap_" not in k)
                and (args.swap == COMPOUND or k != "max_swap")
                and (not args.require_connected or k != "orphan_label")
                and (args.classifier == SPARSE or k not in SPARSE_ARG_NAMES)
                and (args.classifier in NN_CLASSIFIERS or k not in NN_ARG_NAMES | DYNET_ARG_NAMES)
                and k != "passages"]

    def __str__(self):
        self.args.hyperparams = [HyperparamsInitializer(name, **args.vars()) for name, args in self.hyperparams.items()]
        return " ".join(list(self.args.passages) + self.args_str(self.args))



from collections import deque

from semstr.constraints import Constraints, Direction
from semstr.util.amr import LABEL_ATTRIB
from semstr.validation import CONSTRAINTS
from ucca import core, layer0, layer1
from ucca.layer0 import NodeTags
from ucca.layer1 import EdgeTags

from .edge import Edge
from .node import Node
from ..action import Actions
from ..config import Config


class InvalidActionError(AssertionError):
    def __init__(self, *args, is_type=False):
        super().__init__(*args)
        self.is_type = is_type


class State:
    """
    The parser's state, responsible for applying actions and creating the final Passage
    :param passage: a Passage object to get the tokens from, and everything else if training
    """
    def __init__(self, passage):
        self.args = Config().args
        self.constraints = CONSTRAINTS.get(passage.extra.get("format"), Constraints)(implicit=self.args.implicit)
        self.log = []
        self.finished = False
        self.passage = passage
        try:
            l0 = passage.layer(layer0.LAYER_ID)
        except KeyError as e:
            raise IOError("Passage %s is missing layer %s" % (passage.ID, layer0.LAYER_ID)) from e
        try:
            l1 = passage.layer(layer1.LAYER_ID)
        except KeyError:
            l1 = layer1.Layer1(passage)
        self.labeled = any(n.outgoing or n.attrib.get(LABEL_ATTRIB) for n in l1.all)
        self.terminals = [Node(i, orig_node=t, root=passage, text=t.text, paragraph=t.paragraph, tag=t.tag)
                          for i, t in enumerate(l0.all, start=1)]
        self.stack = []
        self.buffer = deque()
        self.nodes = []
        self.heads = set()
        self.need_label = None  # If we are waiting for label_node() to be called, which node is to be labeled by it
        self.root = self.add_node(orig_node=l1.heads[0], is_root=True)  # Root is not in the buffer
        self.stack.append(self.root)
        self.buffer += self.terminals
        self.nodes += self.terminals
        self.actions = []  # History of applied actions
        self.type_validity_cache = {}

    def is_valid_action(self, action):
        """
        :param action: action to check for validity
        :return: is the action (including tag) valid in the current state?
        """
        valid = self.type_validity_cache.get(action.type_id)
        if valid is None:
            try:
                self.check_valid_action(action)
                valid = True
            except InvalidActionError as e:
                valid = False
                if e.is_type:
                    self.type_validity_cache[action.type_id] = valid
        return valid

    def check_valid_action(self, action, message=False):
        """
        Raise InvalidActionError if the action is invalid in the current state
        :param action: action to check for validity
        :param message: whether to add an informative message to the thrown exception
        """
        def _check_possible_node():
            self.check(self.node_ratio() < self.args.max_node_ratio,
                       message and "Non-terminals/terminals ratio: %.3f" % self.args.max_node_ratio, is_type=True)
            for head in self.heads:
                self.check(head.height <= self.args.max_height,
                           message and "Graph height: %d" % self.args.max_height, is_type=True)

        def _check_possible_parent(node, t):
            self.check(node.text is None, message and "Terminals may not have children: %s" % node.text, is_type=True)
            if self.args.constraints and t is not None:
                for rule in self.constraints.tag_rules:
                    violation = rule.violation(node, t, Direction.outgoing, message=message)
                    self.check(violation is None, violation)
                self.check(self.constraints.allow_parent(node, t),
                           message and "%s may not be a '%s' parent (currently %s)" % (
                               node, t, ", ".join(map(str, node.outgoing)) or "childless"))
            self.check(not self.constraints.require_implicit_childless or not node.implicit,
                       message and "Implicit nodes may not have children: %s" % s0, is_type=True)

        def _check_possible_child(node, t):
            self.check(node is not self.root, message and "Root may not have parents", is_type=True)
            if self.args.constraints and t is not None:
                self.check(not t or (node.text is None) != (t == EdgeTags.Terminal),
                           message and "Edge tag must be %s iff child is terminal, but node %s has edge tag %s" %
                           (EdgeTags.Terminal, node, t))
                for rule in self.constraints.tag_rules:
                    violation = rule.violation(node, t, Direction.incoming, message=message)
                    self.check(violation is None, violation)
                self.check(self.constraints.allow_child(node, t),
                           message and "%s may not be a '%s' child (currently %s, %s)" % (
                               node, t, ", ".join(map(str, node.incoming)) or "parentless",
                               ", ".join(map(str, node.outgoing)) or "childless"))
            self.check(self.constraints.possible_multiple_incoming is None or t is None or
                       action.remote or t in self.constraints.possible_multiple_incoming or
                       all(e.remote or e.tag in self.constraints.possible_multiple_incoming for e in node.incoming),
                       message and "Multiple non-remote '%s' parents not allowed for %s" % (t, node))

        def _check_possible_edge(p, c, t):
            _check_possible_parent(p, t)
            _check_possible_child(c, t)
            if self.args.constraints and t is not None:
                if p is self.root:
                    self.check(self.constraints.top_level_allowed is None or not t or
                               t in self.constraints.top_level_allowed, message and "Root may not have %s edges" % t)
                else:
                    self.check(self.constraints.top_level_only is None or
                               t not in self.constraints.top_level_only, message and "Only root may have %s edges" % t)
            self.check(self.constraints.allow_root_terminal_children or p is not self.root or c.text is None,
                       message and "Terminal child '%s' for root" % c, is_type=True)
            if self.constraints.multigraph:  # Nodes may be connected by more than one edge
                edge = Edge(p, c, t, remote=action.remote)
                self.check(self.constraints.allow_edge(edge), message and "Edge not allowed: %s (currently: %s)" % (
                               edge, ", ".join(map(str, p.outgoing)) or "childless"))
            else:  # Simple graph, i.e., no more than one edge between the same pair of nodes
                self.check(c not in p.children, message and "%s is already %s's child" % (c, p), is_type=True)
            self.check(p not in c.descendants, message and "Detected cycle by edge: %s->%s" % (p, c), is_type=True)

        def _check_possible_label():
            self.check(self.args.node_labels, message and "Node labels disabled", is_type=True)
            try:
                node = self.stack[-action.tag]
            except IndexError:
                node = None
            self.check(node is not None, message and "Labeling invalid node %s when stack size is %d" % (
                action.tag, len(self.stack)))
            self.check(not node.labeled, message and "Labeling already-labeled node: %s" % node, is_type=True)
            self.check(node.text is None, message and "Terminals do not have labels: %s" % node, is_type=True)

        if self.args.constraints:
            self.check(self.constraints.allow_action(action, self.actions),
                       message and "Action not allowed: %s " % action + (
                           ("after " + ", ".join("%s" % a for a in self.actions[-3:])) if self.actions else "as first"))
        if action.is_type(Actions.Finish):
            self.check(not self.buffer, "May only finish at the end of the input buffer", is_type=True)
            if self.args.swap:  # Without swap, the oracle may be incapable even of single action
                self.check(self.root.outgoing or all(n is self.root or n.is_linkage or n.text for n in self.nodes),
                           message and "Root has no child at parse end", is_type=True)
            for n in self.nodes:
                self.check(not self.args.require_connected or n is self.root or n.is_linkage or n.text or
                           n.incoming, message and "Non-terminal %s has no parent at parse end" % n, is_type=True)
                self.check(not self.args.node_labels or n.text or n.labeled,
                           message and "Non-terminal %s has no label at parse end" % n, is_type=True)
        else:
            self.check(self.action_ratio() < self.args.max_action_ratio,
                       message and "Actions/terminals ratio: %.3f" % self.args.max_action_ratio, is_type=True)
            if action.is_type(Actions.Shift):
                self.check(self.buffer, message and "Shifting from empty buffer", is_type=True)
            elif action.is_type(Actions.Label):
                _check_possible_label()
            else:   # Unary actions
                self.check(self.stack, message and "%s with empty stack" % action, is_type=True)
                s0 = self.stack[-1]
                if action.is_type(Actions.Reduce):
                    if s0 is self.root:
                        self.check(self.root.labeled or not self.args.node_labels,
                                   message and "Reducing root without label", is_type=True)
                    elif not s0.text:
                        self.check(not self.args.require_connected or s0.is_linkage or s0.incoming,
                                   message and "Reducing parentless non-terminal %s" % s0, is_type=True)
                        self.check(not self.constraints.required_outgoing or
                                   s0.outgoing_tags.intersection((EdgeTags.Terminal, EdgeTags.Punctuation, "")) or
                                   s0.outgoing_tags.intersection(self.constraints.required_outgoing),
                                   message and "Reducing non-terminal %s without %s edge" % (
                                       s0, self.constraints.required_outgoing), is_type=True)
                    self.check(not self.args.node_labels or s0.text or s0.labeled,
                               message and "Reducing non-terminal %s without label" % s0, is_type=True)
                elif action.is_type(Actions.Swap):
                    # A regular swap is possible since the stack has at least two elements;
                    # A compound swap is possible if the stack is longer than the distance
                    distance = action.tag or 1
                    self.check(1 <= distance < len(self.stack), message and "Invalid swap distance: %d" % distance)
                    swapped = self.stack[-distance - 1]
                    # To prevent swap loops: only swap if the nodes are currently in their original order
                    self.check(self.swappable(s0, swapped),
                               message and "Already swapped nodes: %s (swap index %g) <--> %s (swap index %g)"
                               % (swapped, swapped.swap_index, s0, s0.swap_index))
                else:
                    pct = self.get_parent_child_tag(action)
                    self.check(pct, message and "%s with len(stack) = %d" % (action, len(self.stack)), is_type=True)
                    parent, child, tag = pct
                    if parent is None:
                        _check_possible_child(child, tag)
                        _check_possible_node()
                    elif child is None:
                        _check_possible_parent(parent, tag)
                        _check_possible_node()
                    else:  # Binary actions
                        _check_possible_edge(parent, child, tag)

    @staticmethod
    def swappable(right, left):
        return left.swap_index < right.swap_index

    def is_valid_label(self, label):
        """
        :param label: label to check for validity
        :return: is the label valid in the current state?
        """
        try:
            self.check_valid_label(label)
        except InvalidActionError:
            return False
        return True

    def check_valid_label(self, label, message=False):
        if self.args.constraints and label is not None:
            valid = self.constraints.allow_label(self.need_label, label)
            self.check(valid, message and "May not label %s as %s: %s" % (self.need_label, label, valid))

    @staticmethod
    def check(condition, *args, **kwargs):
        if not condition:
            raise InvalidActionError(*args, **kwargs)

    # noinspection PyTypeChecker
    def transition(self, action):
        """
        Main part of the parser: apply action given by oracle or classifier
        :param action: Action object to apply
        """
        action.apply()
        self.log = []
        pct = self.get_parent_child_tag(action)
        if pct:
            parent, child, tag = pct
            if parent is None:
                parent = action.node = self.add_node(orig_node=action.orig_node)
            if child is None:
                child = action.node = self.add_node(orig_node=action.orig_node, implicit=True)
            action.edge = self.add_edge(Edge(parent, child, tag, remote=action.remote))
            if action.node:
                self.buffer.appendleft(action.node)
        elif action.is_type(Actions.Shift):  # Push buffer head to stack; shift buffer
            self.stack.append(self.buffer.popleft())
        elif action.is_type(Actions.Label):
            self.need_label = self.stack[-action.tag]  # The parser is responsible to choose a label and set it
        elif action.is_type(Actions.Reduce):  # Pop stack (no more edges to create with this node)
            self.stack.pop()
        elif action.is_type(Actions.Swap):  # Place second (or more) stack item back on the buffer
            distance = action.tag or 1
            s = slice(-distance - 1, -1)
            self.log.append("%s <--> %s" % (", ".join(map(str, self.stack[s])), self.stack[-1]))
            self.buffer.extendleft(reversed(self.stack[s]))  # extendleft reverses the order
            del self.stack[s]
        elif action.is_type(Actions.Finish):  # Nothing left to do
            self.finished = True
        else:
            raise ValueError("Invalid action: %s" % action)
        if self.args.verify:
            intersection = set(self.stack).intersection(self.buffer)
            assert not intersection, "Stack and buffer overlap: %s" % intersection
        action.index = len(self.actions)
        self.actions.append(action)
        self.type_validity_cache = {}

    def add_node(self, **kwargs):
        """
        Called during parsing to add a new Node (not core.Node) to the temporary representation
        :param kwargs: keyword arguments for Node()
        """
        node = Node(len(self.nodes), swap_index=self.calculate_swap_index(), root=self.passage, **kwargs)
        if self.args.verify:
            assert node not in self.nodes, "Node already exists"
        self.nodes.append(node)
        self.heads.add(node)
        self.log.append("node: %s (swap_index: %g)" % (node, node.swap_index))
        if self.args.use_gold_node_labels:
            self.need_label = node  # Labeled the node as soon as it is created rather than applying a LABEL action
        return node

    def calculate_swap_index(self):
        """
        Update a new node's swap index according to the nodes before and after it.
        Usually the swap index is just the index, i.e., len(self.nodes).
        If the buffer is not empty and its head is not a terminal, it means that it is a non-terminal created before.
        In that case, the buffer head's index will be lower than the new node's index, so the new node's swap index will
        be the arithmetic mean between the previous node (stack top) and the next node (buffer head).
        Then, in the validity check on the SWAP action, we will correctly identify this node as always having appearing
        before the current buffer head. Otherwise, we would prevent swapping them even though it should be valid
        (because they have never been swapped before).
        """
        if self.buffer:
            b0 = self.buffer[0]
            if self.stack and (b0.text is not None or b0.swap_index <= len(self.nodes)):
                s0 = self.stack[-1]
                return (s0.swap_index + b0.swap_index) / 2
        return None

    def add_edge(self, edge):
        edge.add()
        self.heads.discard(edge.child)
        self.log.append("edge: %s" % edge)
        return edge
    
    PARENT_CHILD = (
        ((Actions.LeftEdge, Actions.LeftRemote), (-1, -2)),
        ((Actions.RightEdge, Actions.RightRemote), (-2, -1)),
        ((Actions.Node, Actions.RemoteNode), (None, -1)),
        ((Actions.Implicit,), (-1, None)),
    )

    def get_parent_child_tag(self, action):
        try:
            for types, indices in self.PARENT_CHILD:
                if action.is_type(*types):
                    parent, child = [None if i is None else self.stack[i] for i in indices]
                    break
            else:
                return None
            return parent, child, (EdgeTags.Terminal if child and child.text else
                                   EdgeTags.Punctuation if child and child.children and all(
                                       c.tag == NodeTags.Punct for c in child.children)
                                   else action.tag)  # In unlabeled parsing, keep a valid graph
        except IndexError:
            return None

    def label_node(self, label):
        self.need_label.label = label
        self.need_label.labeled = True
        self.log.append("label: %s" % self.need_label)
        self.type_validity_cache = {}
        self.need_label = None

    def create_passage(self, verify=True, **kwargs):
        """
        Create final passage from temporary representation
        :param verify: fail if this results in an improper passage
        :return: core.Passage created from self.nodes
        """
        Config().print("Creating passage %s from state..." % self.passage.ID, level=2)
        passage = core.Passage(self.passage.ID)
        passage_format = kwargs.get("format") or self.passage.extra.get("format")
        if passage_format:
            passage.extra["format"] = passage_format
        self.passage.layer(layer0.LAYER_ID).copy(passage)
        l0 = passage.layer(layer0.LAYER_ID)
        l1 = layer1.Layer1(passage)
        self.root.node = l1.heads[0]
        if self.args.node_labels:
            self.root.set_node_label()
        if self.labeled:  # We have a reference passage
            self.root.set_node_id()
        Node.attach_nodes(l0, l1, self.nodes, self.labeled, self.args.node_labels, verify)
        return passage

    def node_ratio(self):
        return (len(self.nodes) / len(self.terminals) - 1) if self.terminals else 0

    def action_ratio(self):
        return (len(self.actions) / len(self.terminals)) if self.terminals else 0

    def str(self, sep):
        return "stack: [%-20s]%sbuffer: [%s]" % (" ".join(map(str, self.stack)), sep,
                                                 " ".join(map(str, self.buffer)))

    def __str__(self):
        return self.str(" ")

    def __eq__(self, other):
        return self.stack == other.stack and self.buffer == other.buffer and \
               self.nodes == other.nodes

    def __hash__(self):
        return hash((tuple(self.stack), tuple(self.buffer), tuple(self.nodes)))



import sys
import time
from collections import OrderedDict, Counter, defaultdict

import csv
import json
import numpy as np
import os
import pickle
import pprint as pp
import shutil
from glob import glob
from operator import itemgetter
from tqdm import tqdm

from .labels import Labels

MISSING_VALUE = -1
UNKNOWN_VALUE = 0


class DefaultOrderedDict(OrderedDict, Labels):
    # Source: http://stackoverflow.com/a/6190500/223267
    def __init__(self, default_factory=None, *args, size=None, **kwargs):
        if default_factory is not None and not callable(default_factory):
            raise TypeError("default_factory must be callable")
        Labels.__init__(self, size)
        self._all = []
        OrderedDict.__init__(self, *args, **kwargs)
        self._all = list(self.keys())
        self.default_factory = default_factory

    def __getitem__(self, key):
        try:
            return OrderedDict.__getitem__(self, key)
        except KeyError:
            return self.__missing__(key)

    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = value = self.default_factory()
        return value

    def __reduce__(self):
        args = tuple() if self.default_factory is None else self.default_factory,
        return type(self), args, None, None, iter(self.items())

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        return type(self)(self.default_factory, self)

    def __deepcopy__(self, memo):
        import copy
        return type(self)(self.default_factory, copy.deepcopy(tuple(self.items())))

    def __repr__(self):
        return "%s(%s, %s)" % (type(self), self.default_factory, OrderedDict.__repr__(self))

    def __setitem__(self, key, value, **kwargs):
        super().__setitem__(key, value, **kwargs)
        self._all.append(key)

    @property
    def all(self):
        return self._all

    @all.setter
    def all(self, keys):
        self._all = []
        self.clear()
        for i, key in enumerate(keys):
            self[key] = i


class AutoIncrementDict(DefaultOrderedDict):
    """
    DefaultOrderedDict that returns an auto-incrementing index for new keys
    """
    def __init__(self, size=None, keys=(), d=None, unknown=UNKNOWN_VALUE):
        """
        :param size: maximum size to allow, after which `unknown' will be returned
        :param keys: initial sequence of keys
        :param d: dictionary to initialize from
        :param unknown: value to return for missing keys
        """
        super().__init__(None, sorted(d.items(), key=itemgetter(1)) if d else {}, size=size)
        self.finalized = (size is None)
        self.unknown = self.setdefault(None, unknown)
        for key in keys:
            self.__missing__(key)

    def __missing__(self, key):
        if not self.finalized and len(self) < self.size:
            ret = self[key] = len(self)
            return ret
        return self.unknown

    def first_items(self, n=3):
        return ", ".join(map(str, sorted(self, key=self.get)[:n])) + (", ..." if len(self) > n else "")

    def __str__(self):
        return "{%s}" % self.first_items()

    def __repr__(self):
        return "%s(size=%s, keys=[%s], unknown=%s)" % (type(self).__name__, self.size, self.first_items(), self.unknown)


class UnknownDict(AutoIncrementDict):
    """
    DefaultOrderedDict that has a single default value for missing keys
    """
    def __init__(self, d=None):
        """
        :param d: base dict to initialize by
        """
        super().__init__(size=None, d=d)


class DropoutDict(AutoIncrementDict):
    """
    UnknownDict that sometimes returns the unknown value even for existing keys
    """
    def __init__(self, d=None, dropout=0, size=None, keys=(), min_count=1):
        """
        :param d: base dict to initialize by
        :param dropout: dropout parameter
        :param min_count: minimum number of occurrences for a key before it is actually added to the dict
        """
        super().__init__(size, keys, d=d)
        assert dropout >= 0, "Dropout value must be >= 0, but given %f" % dropout
        self.dropout, self.counts, self.min_count = (d.dropout, d.counts, d.min_count) \
            if d is not None and isinstance(d, DropoutDict) else (dropout, Counter(), min_count)

    def __getitem__(self, item):
        if item is not None:
            self.counts[item] += 1
            count = self.counts[item]
            if count < self.min_count or self.dropout and self.dropout/(count+self.dropout) > np.random.random_sample():
                item = None
        return super().__getitem__(item)


class KeyBasedDefaultDict(defaultdict):
    def __missing__(self, key):
        if self.default_factory is None:
            raise KeyError(key)
        self[key] = self.default_factory(key)
        return self[key]


def remove_existing(*filenames):
    for filename in filenames:
        try:
            shutil.copy2(filename, filename + "~")
            os.remove(filename)
            print("Removed existing '%s'." % filename)
        except OSError:
            pass


def remove_backup(*filenames):
    for filename in filenames:
        for backup in glob(filename + "*~"):
            try:
                os.remove(backup)
            except OSError:
                pass


def save_dict(filename, d):
    """
    Save dictionary to Pickle file
    :param filename: file to write to
    :param d: dictionary to save
    """
    remove_existing(filename)
    sys.setrecursionlimit(2000)
    print("Saving to '%s'... " % filename, end="", flush=True)
    started = time.time()
    with open(filename, "wb") as h:
        try:
            pickle.dump(d, h, protocol=pickle.HIGHEST_PROTOCOL)
        except RecursionError as e:
            raise IOError("Failed dumping dictionary:\n" + pp.pformat(d, compact=True)) from e
    print("Done (%.3fs)." % (time.time() - started))


def load_dict(filename):
    """
    Load dictionary from Pickle file
    :param filename: file to read from
    """

    def try_load(*names):
        exception = None
        for f in names:
            # noinspection PyBroadException
            try:
                with open(f, "rb") as h:
                    return pickle.load(h)
            except FileNotFoundError as e:
                exception = e
        if exception is not None:
            raise FileNotFoundError("File not found: '%s'" % "', '".join(names)) from exception

    print("Loading from '%s'... " % filename, end="", flush=True)
    started = time.time()
    d = try_load(filename, os.path.splitext(filename)[0])
    print("Done (%.3fs)." % (time.time() - started))
    return d


def jsonify(o):
    try:
        return o.__dict__
    except AttributeError:
        return o.tolist()


def save_json(filename, d):
    """
    Save dictionary to JSON file
    :param filename: file to write to
    :param d: dictionary to save
    """
    remove_existing(filename)
    print("Saving to '%s'." % filename)
    with open(filename, "w") as h:
        json.dump(d, h, default=jsonify)


def load_json(filename):
    """
    Load dictionary from JSON file
    :param filename: file to read from
    """
    print("Loading from '%s'." % filename)
    with open(filename, "r") as h:
        d = json.load(h)
    return d


class Lexeme:
    def __init__(self, index, text):
        self.index = self.orth = index
        self.text = self.orth_ = text


class Strings:
    def __init__(self, vocab):
        self.vocab = vocab

    def __getitem__(self, item):
        lex = self.vocab[item]
        return lex.index if isinstance(item, str) else lex.text


class Vocab(dict):
    def __init__(self, tuples):
        super().__init__()
        for k, v in tuples:
            self[int(k)] = self[v] = Lexeme(int(k), v)
        self.strings = Strings(self)


class IdentityVocab(Vocab):
    def __init__(self):
        super().__init__(())

    def __contains__(self, item):
        return True

    def __getitem__(self, item):
        return Lexeme(item, item)


def load_enum(filename):
    if filename == "-":
        return IdentityVocab()
    else:
        with open(filename, encoding="utf-8") as f:
            return Vocab(tqdm(csv.reader(f), desc="Loading '%s'" % filename, file=sys.stdout, unit=" rows"))



from collections import OrderedDict

import numpy as np
from ucca.textutil import get_word_vectors

from ..config import Config
from ..labels import Labels
from ..model_util import DropoutDict


class FeatureParameters(Labels):
    def __init__(self, suffix, dim, size, dropout=0, updated=True, num=1, init=None, data=None, indexed=False,
                 copy_from=None, filename=None, min_count=1, enabled=True, node_dropout=0, vocab=None,
                 lang_specific=False):
        """
        :param suffix: one-character title for feature
        :param dim: vector dimension or, filename to load vectors from, or Word2Vec object
        :param size: maximum number of distinct values
        :param dropout: value of dropout parameter to use during training
        :param updated: whether the feature is learned (otherwise kept constant)
        :param num: how many such features exist per step
        :param init: array of values to use as initial value of embedding matrix
        :param data: DefaultOrderedDict of raw value to running numerical representation, or embedding matrix
        :param indexed: whether the feature is to be used as index into initialized values (otherwise used directly)
        :param copy_from: suffix of other parameter to copy values from instead of extracting them directly
        :param filename: name of file to load data from
        :param min_count: minimum number of occurrences for a feature value before it is actually added
        :param enabled: whether to actually use this parameter in feature extraction
        :param node_dropout: probability to drop whole node in feature extraction
        :param vocab: name of file to load mapping of integer ID to word form (to avoid loading spaCy)
        :param lang_specific: whether the feature params should be separate per language
        """
        super().__init__(size)
        self.suffix = suffix
        self.dim = dim
        self.size = size
        self.dropout = dropout
        self.updated = updated
        self.num = num
        self.init = init
        self.data = data
        self.indexed = indexed
        self.copy_from = copy_from
        self.filename = filename
        self.min_count = min_count
        self.enabled = enabled
        self.node_dropout = node_dropout
        self.vocab = vocab
        self.lang_specific = lang_specific

    def __repr__(self):
        return type(self).__name__ + "(" + ", ".join(
            map(str, (self.suffix, self.dim, self.size, self.dropout, self.updated, self.num, self.init, self.data,
                      self.indexed, self.copy_from, self.filename, self.min_count, self.enabled, self.node_dropout,
                      self.vocab, self.lang_specific))) + ")"

    def __eq__(self, other):
        return self.suffix == other.suffix and self.dim == other.dim and self.size == other.size and \
               self.dropout == other.dropout and self.updated == other.updated and self.num == other.num and \
               self.indexed == other.indexed and self.min_count == other.min_count and \
               self.numeric == other.numeric and self.node_dropout == other.node_dropout and \
               self.lang_specific == other.lang_specific

    def __hash__(self):
        return hash(self.suffix)

    def init_data(self):
        if self.data is None and not self.numeric:
            keys = ()
            if self.dim and self.external:
                vectors = self.word_vectors()
                keys = vectors.keys()
                self.init = np.array(list(vectors.values()))
            self.data = DropoutDict(size=self.size, keys=keys, dropout=self.dropout, min_count=self.min_count)

    def word_vectors(self):
        lang = Config().args.lang
        vectors, self.dim = get_word_vectors(self.dim, self.size, self.filename, Config().vocab(self.vocab) or lang)
        if self.size is not None:
            assert len(vectors) <= self.size, "Wrong number of loaded vectors: %d > %d" % (len(vectors), self.size)
        assert vectors, "Cannot load word vectors. Install using `python -m spacy download %s` or choose a file " \
                        "using the --word-vectors option." % lang
        self.size = len(vectors)
        return vectors

    @property
    def all(self):
        return self.data.all

    @all.setter
    def all(self, labels):
        self.data.all = labels

    @property
    def numeric(self):
        return False

    @property
    def prop(self):
        return self.copy_from or self.suffix

    @property
    def external(self):
        return self.copy_from is not None

    @staticmethod
    def copy(params, copy_dict=dict, copy_init=True, order=None):
        return OrderedDict((key, param.copy_with_data(copy_dict, copy_init)) for key, param in
                           (params.items() if order is None else sorted(params.items(), key=lambda x:
                            -1 if x[0] == NumericFeatureParameters.SUFFIX else order.index(x[0][0]))))

    def copy_with_data(self, copy_dict, copy_init):
        data = None if self.data is None else copy_dict(self.data)
        if hasattr(data, "size"):  # It may be an UnknownDict but we still want it to know its size
            data.size = self.size
        return FeatureParameters(suffix=self.suffix, dim=self.dim, size=self.size, dropout=self.dropout,
                                 updated=self.updated, num=self.num, init=self.init if copy_init else None, data=data,
                                 indexed=self.indexed, copy_from=self.copy_from, filename=self.filename,
                                 min_count=self.min_count, enabled=self.enabled,
                                 node_dropout=getattr(self, "node_dropout", 0),
                                 vocab=getattr(self, "vocab", None),
                                 lang_specific=getattr(self, "lang_specific", False))

    def unfinalize(self):
        self.data = DropoutDict(self.data, size=self.size, dropout=self.dropout, min_count=self.min_count)


class NumericFeatureParameters(FeatureParameters):
    SUFFIX = "numeric"

    def __init__(self, num, node_dropout=0):
        super().__init__(NumericFeatureParameters.SUFFIX, 1, None, num=num, node_dropout=node_dropout)

    def __repr__(self):
        return "%s(%d)" % (type(self).__name__, self.num)

    @property
    def numeric(self):
        return True

    def copy_with_data(self, copy_dict, copy_init):
        return NumericFeatureParameters(self.num, node_dropout=getattr(self, "node_dropout", 0))



from configargparse import ArgParser
from ucca.textutil import get_vocab

from tupa.model_util import save_json
from tupa.scripts.export import load_model


def decode(vocab, value):
    if isinstance(value, int):
        try:
            return vocab.strings[value]
        except KeyError:
            pass
    return value


def main():
    argparser = ArgParser(description="Load TUPA model and save the features enumeration as a text JSON file.")
    argparser.add_argument("models", nargs="+", help="model file basename(s) to load")
    argparser.add_argument("-s", "--suffix", default=".enum.json", help="filename suffix to append")
    argparser.add_argument("-l", "--lang", help="use spaCy model to decode numeric IDs")
    args = argparser.parse_args()
    for filename in args.models:
        model = load_model(filename)
        params = model.feature_extractor.params
        if args.lang:
            vocab = get_vocab(lang=args.lang)
            for param in params.values():
                if param.data:
                    param.data = [decode(vocab, v) for v in sorted(param.data, key=param.data.get)]
        save_json(model.filename + args.suffix, params)


if __name__ == "__main__":
    main()



from .config import Config, COMPOUND
from .labels import Labels


class Action(dict):
    type_to_id = {}

    def __init__(self, action_type, tag=None, orig_edge=None, orig_node=None, oracle=None, id_=None):
        self.type = action_type  # String
        self.tag = tag  # Usually the tag of the created edge; but if COMPOUND_SWAP, the distance
        self.orig_node = orig_node  # Node created by this action, if any (during training)
        self.orig_edge = orig_edge  # Edge created by this action, if any (during training)
        self.node = None  # Will be set by State when the node created by this action is known
        self.edge = None  # Will be set by State when the edge created by this action is known
        self.oracle = oracle  # Reference to oracle, to inform it of actually created nodes/edges
        self.index = None  # Index of this action in history

        self.type_id = Action.type_to_id.get(self.type)  # Allocate ID for fast comparison
        if self.type_id is None:
            self.type_id = len(Action.type_to_id)
            Action.type_to_id[self.type] = self.type_id
        self.id = id_
        super().__init__(action_type=self.type, tag=self.tag)

    def is_type(self, *others):
        return self.type_id in (o.type_id for o in others)

    def apply(self):
        if self.oracle is not None:
            self.oracle.remove(self.orig_edge, self.orig_node)

    def __repr__(self):
        return Action.__name__ + "(" + ", ".join(map(str, filter(None, (self.type, self.tag)))) + ")"

    def __str__(self):
        s = self.type
        if self.tag:
            s += "-%s" % self.tag
        return s

    def __eq__(self, other):
        return self.id == other.id

    def __hash__(self):
        return hash(self.id)

    def __call__(self, *args, **kwargs):
        return Action(self.type, *args, **kwargs)

    @property
    def remote(self):
        return self.is_type(Actions.RemoteNode, Actions.LeftRemote, Actions.RightRemote)

    @property
    def is_swap(self):
        return self.is_type(Actions.Swap)


class Actions(Labels):
    Shift = Action("SHIFT")
    Node = Action("NODE")
    RemoteNode = Action("REMOTE-NODE")
    Implicit = Action("IMPLICIT")
    Label = Action("LABEL")
    Reduce = Action("REDUCE")
    LeftEdge = Action("LEFT-EDGE")
    RightEdge = Action("RIGHT-EDGE")
    LeftRemote = Action("LEFT-REMOTE")
    RightRemote = Action("RIGHT-REMOTE")
    Swap = Action("SWAP")
    Finish = Action("FINISH")

    def __init__(self, actions=None, size=None):
        super().__init__(size=size)
        self._all = None
        self._ids = None
        if actions is not None:
            self.all = actions

    def init(self):
        # edge and node action will be created as they are returned by the oracle
        args = Config().args
        self.all = [Actions.Reduce, Actions.Shift, Actions.Finish] + \
            (list(map(Actions.Swap, range(1, args.max_swap))) if args.swap == COMPOUND else
             [Actions.Swap] if args.swap else [])

    @property
    def all(self):
        if self._all is None:
            self.init()
        return self._all

    @all.setter
    def all(self, actions):
        self._all = [Action(**a) if isinstance(a, dict) else a for a in actions]
        self._ids = {(action.type_id, action.tag): i for i, action in enumerate(self._all)}
        for action in self._all:
            self.generate_id(action)

    @property
    def ids(self):
        if self._all is None:
            self.init()
        return self._ids

    def generate_id(self, action, create=True):
        if action.id is None:
            key = (action.type_id, action.tag)
            action.id = self.ids.get(key)
            if create and action.id is None:  # New action, add to list
                # noinspection PyTypeChecker
                action.id = len(self.all)
                self.all.append(action(tag=action.tag, id_=action.id))
                self.ids[key] = action.id



from collections import OrderedDict

import matplotlib
import numpy as np

matplotlib.use("Agg")
import matplotlib.pyplot as plt
from configargparse import ArgParser
from scipy.misc import imresize
from matplotlib.ticker import MaxNLocator
from tupa.scripts.export import load_model
from tqdm import tqdm
import re

np.seterr("raise")

REPLACEMENTS = (
    ("axes_", ""),
    ("_", " "),
)


def smooth(x, s=100000):
    if len(x.shape) == 1:
        x = x.reshape((x.shape[0], -1))
    return imresize(x, size=(min(x.shape[0], s), min(x.shape[1], s)))


def visualize(model, filename):
    values = model.all_params()
    cols = OrderedDict()
    for key, value in values.items():
        if isinstance(value, np.ndarray):  # TODO group by prefix, showing W and b side by side
            cols.setdefault(key[:-2] + key[-1] if len(key) > 1 and key[-2] in "bW" else key, []).append((key, value))
    _, axes = plt.subplots(2, len(cols), figsize=(5*len(cols), 10))  # TODO https://stackoverflow.com/a/13784887/223267
    plt.tight_layout()
    for j, col in enumerate(tqdm(cols.values(), unit="param", desc=filename)):
        for i in range(2):
            axis = axes[i, j] if len(values) > 1 else axes
            if len(col) <= i:
                plt.delaxes(axis)
            else:
                plt.sca(axis)
                key, value = col[i]
                plt.colorbar(plt.pcolormesh(smooth(value)))
                for pattern, repl in REPLACEMENTS:  # TODO map 0123->ifoc
                    key = re.sub(pattern, repl, key)
                plt.title(key)
                for axis in (plt.gca().xaxis, plt.gca().yaxis):
                    axis.set_major_locator(MaxNLocator(integer=True))
    output_file = filename + ".png"
    plt.savefig(output_file)
    plt.clf()
    print("Saved '%s'." % output_file)


def main():
    argparser = ArgParser(description="Load TUPA model and visualize, saving to .png file.")
    argparser.add_argument("models", nargs="+", help="model file basename(s) to load")
    args = argparser.parse_args()
    for filename in args.models:
        model = load_model(filename)
        visualize(model, filename)


if __name__ == "__main__":
    main()



import re
from ucca import layer0
from ucca.layer1 import EdgeTags
from ucca.textutil import Attr

from tupa.config import Config, FEATURE_PROPERTIES

FEATURE_ELEMENT_PATTERN = re.compile(r"([sba])(\d)([lrLR]*)([%s]*)" % FEATURE_PROPERTIES)
FEATURE_TEMPLATE_PATTERN = re.compile(r"^(%s)+$" % FEATURE_ELEMENT_PATTERN.pattern)
NON_NUMERIC = "wmtudencpAT#^$"


class FeatureTemplate:
    """
    A feature template in parsed form, ready to be used for value calculation
    """

    def __init__(self, name, elements):
        """
        :param name: name of the feature in the short-hand form, to be used for the dictionary
        :param elements: collection of FeatureElement objects that represent the actual feature
        """
        self.name = name
        self.elements = elements
        for prev, elem in zip(self.elements[:-1], self.elements[1:]):
            if not prev.properties:
                assert prev.previous is None, "More than one consecutive empty element: %s%s" % (prev.previous, prev)
                elem.previous = prev

    def __str__(self):
        return self.name

    def extract(self, state, default=None, indexed=(), as_tuples=False, node_dropout=0, hierarchical=False):
        try:
            return [value for element in self.elements for value in element.extract(state, default, indexed, as_tuples,
                                                                                    node_dropout=node_dropout,
                                                                                    hierarchical=hierarchical)]
        except ValueError:
            return None


class FeatureTemplateElement:
    """
    One element in the values of a feature, e.g. from one node
    """

    def __init__(self, source, index, relatives, properties, omit_features=None):
        """
        :param source: where to take the data from:
                           s: stack nodes
                           b: buffer nodes
                           a: past actions
        :param index: non-negative integer, the index of the element in the stack, buffer or list
                           of past actions (in the case of stack and actions, indexing from the end)
        :param relatives: string in [lrLR]*, to select a descendant/parent of the node instead:
                           l: leftmost child
                           r: rightmost child
                           L: leftmost parent
                           R: rightmost parent
        :param properties: the actual values to choose, if available (else omit feature), out of:
                           w: node text
                           m: node lemma
                           t: node fine POS tag
                           u: node coarse/universal POS tag
                           d: node dependency relation
                           h: node height
                           e: tag of first incoming edge / action tag
                           n: node label
                           c: node label category suffix
                           p: unique separator punctuation between nodes
                           q: count of any separator punctuation between nodes
                           x: numeric value of gap type
                           y: sum of gap lengths
                           A: action type label
                           P: number of parents
                           C: number of children
                           I: number of implicit children
                           E: number of remote children
                           M: number of remote parents
                           N: numeric value of named entity IOB
                           T: named entity type
                           #: word shape
                           ^: word prefix (one character)
                           $: word suffix (three characters)
                           If empty,
                             If the next node comes with the "x" property, the value will be 1 if there is an edge from
                             this node to the next one in the template, or 0 otherwise.
                             If the next node comes with the "e" property, the edge with this node will be considered.
                             If the next node comes with the "d" property, the value will be the dependency distance
                             between the head terminals of the nodes.
        """
        self.source = source
        self.index = int(index)
        self.relatives = relatives
        self.properties = properties.translate(str.maketrans("", "", omit_features)) if omit_features else properties
        self.node = self.previous = None
        self.getters = [prop_getter(prop, self.source) for prop in self.properties]

    def __str__(self):
        return self.str + self.properties

    @property
    def str(self):
        return (self.previous.str if self.previous else "") + self.source + str(self.index) + self.relatives

    def __eq__(self, other):
        return self.source == other.source and self.index == other.index and self.relatives == other.relatives

    def set_node(self, state, node_dropout=0):
        self.node = None
        if state is None or node_dropout and node_dropout > Config().random.random_sample():
            return
        try:
            if self.source == "s":
                self.node = state.stack[-1 - self.index]
            elif self.source == "b":
                self.node = state.buffer[self.index]
            else:  # source == "a"
                self.node = state.actions[-1 - self.index]
            for relative in self.relatives:
                nodes = self.node.parents if relative.isupper() else self.node.children
                if relative.lower() == "r":
                    if len(nodes) == 1:
                        raise ValueError("Avoiding identical right and left relatives")
                    self.node = nodes[-1]
                else:  # relative.lower() == "l"
                    self.node = nodes[0]
        except (IndexError, TypeError, AttributeError, IndexError, ValueError):
            if Config().args.missing_node_features or node_dropout and node_dropout > Config().random.random_sample():
                self.node = None

    def extract(self, state, default, indexed, as_tuples, node_dropout=0, hierarchical=False):
        self.set_node(state, node_dropout=node_dropout)
        for prop, getter in zip(self.properties, self.getters):
            if indexed and not self.is_numeric(prop):
                if prop == indexed[0]:
                    getter = NODE_PROP_GETTERS["j" if hierarchical else "i"]
                elif prop in indexed[1:]:
                    continue
            value = self.get_prop(state, prop, getter, default)
            yield (self, prop, value) if as_tuples else value

    def get_prop(self, state, prop, getter, default):
        value = calc(self.node, state, prop, getter, self.previous)
        if value is None:
            if default is None:
                raise ValueError("Value does not exist, and no default given")
            value = default
        return value

    def is_numeric(self, prop):
        return bool(prop not in NON_NUMERIC or (prop == "d" and self.previous))


class FeatureExtractor:
    """
    Object to extract features from the parser state to be used in action classification
    """

    def __init__(self, feature_templates=(), params=None, omit_features=None):
        assert all(FEATURE_TEMPLATE_PATTERN.match(f) for f in feature_templates), \
            "Features do not match pattern: " + ", ".join(
                f for f in feature_templates if not FEATURE_TEMPLATE_PATTERN.match(f))
        # convert the list of features textual descriptions to the actual fields
        self.feature_templates = [
            FeatureTemplate(feature_name, tuple(FeatureTemplateElement(*m.group(1, 2, 3, 4), omit_features)
                                                for m in re.finditer(FEATURE_ELEMENT_PATTERN, feature_name)))
            for feature_name in feature_templates]
        self.params = {} if params is None else params
        self.omit_features = omit_features

    def extract_features(self, state):
        """
        Calculate feature values according to current state
        :param state: current state of the parser
        """
        raise NotImplementedError()

    def init_features(self, state):
        """
        Calculate feature values for initial state
        :param state: initial state of the parser
        """
        pass

    def init_param(self, key):
        pass

    def finalize(self):
        return self

    def unfinalize(self):
        pass

    def save(self, filename, save_init=True):
        pass

    def load(self, filename, order=None):
        pass

    def all_features(self):
        return sorted(list(map(str, self.feature_templates)))


EDGE_PRIORITY = {tag: i for i, tag in enumerate((
    EdgeTags.Center,
    EdgeTags.Connector,
    EdgeTags.ParallelScene,
    EdgeTags.Process,
    EdgeTags.State,
    EdgeTags.Participant,
    EdgeTags.Adverbial,
    EdgeTags.Time,
    EdgeTags.Elaborator,
    EdgeTags.Relator,
    EdgeTags.Function,
    EdgeTags.Linker,
    EdgeTags.LinkRelation,
    EdgeTags.LinkArgument,
    EdgeTags.Ground,
    EdgeTags.Terminal,
    EdgeTags.Punctuation,
))}


def head_terminal(node, *_):
    return head_terminal_height(node)


def height(node, *_):
    return head_terminal_height(node, return_height=True)


def static_vars(**kwargs):
    def decorate(func):
        for k, v in kwargs.items():
            setattr(func, k, v)
        return func
    return decorate


MAX_HEIGHT = 30


@static_vars(node=None, head_terminal=None, height=None)
def head_terminal_height(node, return_height=False):
    if node is not head_terminal_height.node:
        head_terminal_height.node = head_terminal_height.head_terminal = node
        head_terminal_height.height = 0
        while head_terminal_height.head_terminal.text is None:  # Not a terminal
            edges = [edge for edge in node.outgoing if not edge.remote and not edge.child.implicit]
            if not edges or head_terminal_height.height > MAX_HEIGHT:
                head_terminal_height.head_terminal = head_terminal_height.height = None
                break
            head_terminal_height.head_terminal = min(edges, key=lambda edge: EDGE_PRIORITY.get(edge.tag, 0)).child
            head_terminal_height.height += 1
    return head_terminal_height.height if return_height else head_terminal_height.head_terminal


def has_gaps(node, *_):  # Possibly the same as FoundationalNode.discontiguous
    return any(length > 0 for length in gap_lengths(node))


def gap_length_sum(node, *_):
    return sum(gap_lengths(node))


def gap_lengths(node, *_):
    terminals = node.get_terminals()
    return (t1.index - t2.index - 1 for (t1, t2) in zip(terminals[1:], terminals[:-1]))


def gap_type(node, *_):
    if node.text is None:  # Not a terminal
        if has_gaps(node):
            return 1  # Pass
        if any(child.text is None and has_gaps(child) for child in node.children):
            return 2  # Source
    return 0  # None


def dep_distance(node1, node2, *_):
    terminals = list(map(head_terminal, (node1, node2)))
    for v in 1, -1:
        t1, t2 = terminals[::v]
        if t1.index + t1.tok[Attr.HEAD.value] == t2.index:
            return v
    return None


def get_punctuation(nodes, terminals):
    if len(nodes) < 2:
        return None
    t0, t1 = sorted([head_terminal(node) for node in nodes], key=lambda t: t.index)
    return [t for t in terminals[t0.index:t1.index - 1] if t.tag == layer0.NodeTags.Punct]


ACTION_PROP_GETTERS = {
    "A": lambda a, *_: a.type,
    "e": lambda a, *_: a.tag if isinstance(a.tag, str) or Config().args.missing_node_features else None,  # Swap, Label
}


NODE_PROP_GETTERS = {
    "w": lambda node, *_: head_terminal(node).tok[Attr.ORTH.value],
    "m": lambda node, *_: head_terminal(node).tok[Attr.LEMMA.value],
    "t": lambda node, *_: head_terminal(node).tok[Attr.TAG.value],
    "u": lambda node, *_: head_terminal(node).tok[Attr.POS.value],
    "d": lambda node, prev, binary: dep_distance(prev, node) if binary else head_terminal(node).tok[Attr.DEP.value],
    "h": height,
    "i": lambda node, *_: head_terminal(node).index - 1,
    "j": lambda node, *_: node.index,
    "e": lambda node, prev, binary: next(e.tag for e in node.incoming if not binary or e.parent == prev),
    "n": lambda node, *_: node.label,
    "c": lambda node, *_: node.category,
    "x": lambda node, prev, binary: int(prev in node.parents) if binary else gap_type(node),
    "y": gap_length_sum,
    "P": lambda node, *_: len(node.incoming),
    "C": lambda node, *_: len(node.outgoing),
    "I": lambda node, *_: sum(1 for n in node.children if n.implicit),
    "E": lambda node, *_: sum(1 for e in node.outgoing if e.remote),
    "M": lambda node, *_: sum(1 for e in node.incoming if e.remote),
    "N": lambda node, *_: head_terminal(node).tok[Attr.ENT_IOB.value],
    "T": lambda node, *_: head_terminal(node).tok[Attr.ENT_TYPE.value],
    "#": lambda node, *_: head_terminal(node).tok[Attr.SHAPE.value],
    "^": lambda node, *_: head_terminal(node).tok[Attr.PREFIX.value],
    "$": lambda node, *_: head_terminal(node).tok[Attr.SUFFIX.value],
}


SEP_PROP_GETTERS = {
    "p": lambda nodes, terminals: get_punctuation(nodes, terminals)[0].tok[Attr.ORTH.value],
    "q": lambda nodes, terminals: len(get_punctuation(nodes, terminals)),
}


def prop_getter(prop, source=None):
    return (ACTION_PROP_GETTERS if source == "a" else SEP_PROP_GETTERS if prop in "pq" else NODE_PROP_GETTERS)[prop]


def calc(node, state, prop=None, getter=None, prev=None):
    if node is None:
        return None
    if getter is None:
        getter = prop_getter(prop)
    try:
        if prop in "pq":
            return getter(state.stack[-1:-3:-1], state.terminals)
        return getter(node, None if prev is None else prev.node, prev is not None)
    except (TypeError, AttributeError, IndexError, StopIteration):
        return None



import os
from base64 import b64encode
from io import BytesIO
from urllib.parse import quote
from xml.etree.ElementTree import fromstring, tostring

import flask_assets
import jinja2
import matplotlib
from flask import Flask, render_template, Response, request
from flask_compress import Compress
from ucca import layer1
from ucca.convert import from_text, to_standard, from_standard
from ucca.textutil import indent_xml
from ucca.visualization import draw
from webassets import Environment as AssetsEnvironment
from webassets.ext.jinja2 import AssetsExtension

from semstr.convert import TO_FORMAT
from tupa.parse import Parser

matplotlib.use("Agg")
import matplotlib.pyplot as plt
from matplotlib.backends.backend_agg import FigureCanvasAgg

SCRIPT_DIR = os.path.dirname(__file__)

app = Flask(__name__)
assets = flask_assets.Environment()
assets.init_app(app)
assets_env = AssetsEnvironment("./static/", "/static")
jinja_environment = jinja2.Environment(
    autoescape=True,
    loader=jinja2.FileSystemLoader(os.path.join(SCRIPT_DIR, "templates")),
    extensions=[AssetsExtension])
jinja_environment.assets_environment = assets_env
Compress(app)

app.parser = None
PARSER_MODEL = os.getenv("PARSER_MODEL", os.path.join(SCRIPT_DIR, "..", "models/bilstm"))


def get_parser():
    if app.parser is None:
        print("Initializing parser...")
        print("PARSER_MODEL=" + PARSER_MODEL)
        app.parser = Parser(PARSER_MODEL)
    return app.parser


@app.route("/")
def parser_demo():
    return render_template("demo.html")


@app.route("/parse", methods=["POST"])
def parse():
    text = request.values["input"]
    print("Parsing text: '%s'" % text)
    in_passage = next(from_text(text))
    out_passage = next(get_parser().parse(in_passage))[0]
    root = to_standard(out_passage)
    xml = tostring(root).decode()
    return Response(indent_xml(xml), headers={"Content-Type": "xml/application"})


@app.route("/visualize", methods=["POST"])
def visualize():
    xml = request.get_data()
    passage = from_standard(fromstring(xml))
    print("Visualizing passage %s: %s" % (passage.ID, passage.layer(layer1.LAYER_ID).heads[0]))
    fig = plt.figure()
    canvas = FigureCanvasAgg(fig)
    draw(passage)
    image = BytesIO()
    canvas.print_png(image)
    data = b64encode(image.getvalue()).decode()
    plt.close()
    return Response(quote(data.rstrip("\n")))

CONTENT_TYPES = {"xml": "xml/application", "json": "application/json"}


@app.route("/download", methods=["POST"])
def download():
    xml = request.values["input"]
    out_format = request.values["format"]
    print("Converting to " + out_format)
    out = xml if out_format == "xml" else "\n".join(TO_FORMAT[out_format](from_standard(fromstring(xml))))
    return Response(out, headers={"Content-Type": CONTENT_TYPES.get(out_format, "text/plain")})


session_opts = {
    "session.type": "file",
    "session.cookie_expires": 60 * 24 * 60 * 2,  # two days in seconds
    "session.data_dir": "./data",
    "session.auto": True
}

if __name__ == "__main__":
    app.run(debug=True, host=os.getenv("IP", "ucca"), port=int(os.getenv("PORT", 5001)))



import time
from collections import defaultdict
from itertools import repeat

import numpy as np

from tupa.model_util import KeyBasedDefaultDict, save_dict, load_dict
from ..classifier import Classifier


class FeatureWeights:
    """
    The weights for one feature, for all labels
    """
    def __init__(self, num_labels=None, weights=None):
        if num_labels is not None and weights is None:
            self.weights = np.zeros(num_labels, dtype=float)  # 0.01 * np.random.randn(num_labels)
            self._totals = np.zeros(num_labels, dtype=float)
            self._last_update = np.zeros(num_labels, dtype=int)
            self.update_count = 0
        else:
            self.weights = weights

    def update(self, label, value, update_index):
        """
        Add a value to the entry of the given label
        :param label: label to index by
        :param value: value to add
        :param update_index: which update this is (for averaging)
        """
        self.update_count += 1
        self._update_totals(label, update_index)
        self.weights[label] += value

    def finalize(self, update_index, average):
        """
        Average weights over all updates, and keep only true label columns
        :param update_index: number of updates to average over
        :param average: whether to really average the weights or just return them as they are now
        :return new Weights object with the weights averaged and only selected indices remaining
        """
        self._update_totals(None, update_index)
        weights = self._totals / update_index if average else self.weights
        return FeatureWeights(weights=weights)

    def _update_totals(self, label, update_index):
        self._totals[label] += self.weights[label] * (update_index - self._last_update[label])
        self._last_update[label] = update_index

    def resize(self, num_labels):
        self.weights.resize(num_labels, refcheck=False)
        self._totals.resize(num_labels, refcheck=False)
        self._last_update.resize(num_labels, refcheck=False)


class FeatureWeightsCreator:
    def __init__(self, perceptron, axis):
        self.perceptron = perceptron
        self.axis = axis

    def create(self):
        return FeatureWeights(self.perceptron.num_labels[self.axis])


class SparsePerceptron(Classifier):
    """
    Multi-class averaged perceptron with min-update for sparse features.
    Keeps weights in a dictionary by feature name, allowing adding new features on-the-fly.
    Also allows adding new labels on-the-fly.
    Expects features from SparseFeatureExtractor.
    """

    def __init__(self, *args, epoch=0, **kwargs):
        """
        Create a new untrained SparsePerceptron or copy the weights from an existing one
        :param labels: tuple of lists of labels that can be updated later to add new labels
        :param min_update: minimum number of updates to a feature required for consideration
        :param model: if given, copy the weights (from a trained model)
        """
        super().__init__(*args, **kwargs)
        model = KeyBasedDefaultDict(self.create_axis_weights)
        if self.is_frozen:
            for axis, feature_weights in self.model.items():
                model[axis].update(feature_weights)
        self.model = model
        self.min_update = self.config.args.min_update  # Minimum number of updates for a feature to be used in scoring
        self.dropped = set()  # Features that did not get min_updates after a full epoch
        self.initial_learning_rate = self.learning_rate if self.learning_rate else 1.0
        self.epoch = epoch
        self.update_learning_rate()

    @property
    def input_dim(self):
        return {a: len(m) for a, m in self.model.items()}

    def update_learning_rate(self):
        self.learning_rate = self.initial_learning_rate / (1.0 + self.epoch * self.learning_rate_decay)

    def create_axis_weights(self, axis):
        return defaultdict(FeatureWeightsCreator(self, axis).create)

    def copy_model(self):
        return {a: dict(m) for a, m in self.model.items()}

    def update_model(self, model):
        for axis, feature_weights in model.items():
            self.model[axis].update(feature_weights)

    def score(self, features, axis):
        """
        Calculate score for each label
        :param features: extracted feature values, in the form of a dict (name -> value)
        :param axis: axis of the label we are predicting
        :return: array with score for each label
        """
        super().score(features, axis)
        scores = np.zeros(self.num_labels[axis])
        model = self.model[axis]
        for feature, value in features.items():
            if not value:
                continue
            weights = model.get(feature)
            if weights is not None:
                if self.is_frozen or weights.update_count >= self.min_update:
                    scores += value * weights.weights
        return scores

    def update(self, features, axis, pred, true, importance=None):
        """
        Update classifier weights according to predicted and true labels
        :param features: extracted feature values, in the form of a dict (name: value)
        :param axis: axis of the label we are predicting
        :param pred: label predicted by the classifier (non-negative integer bounded by num_labels[axis])
        :param true: true labels (non-negative integers bounded by num_labels[axis])
        :param importance: how much to scale the update for the weight update for each true label
        """
        super().update(features, axis, pred, true, importance)
        self.updates += 1
        model = self.model[axis]
        for feature, value in features.items():
            if not value or feature in self.dropped:
                continue
            weights = model[feature]
            for t, i in zip(true, importance or repeat(1)):
                weights.update(t, i * self.learning_rate * value, self.updates)
            weights.update(pred, -self.learning_rate * value, self.updates)

    def resize(self):
        for axis, model in self.model.items():
            num_labels = self.num_labels[axis]
            for weights in model.values():
                weights.resize(num_labels)

    def finalize(self, finished_epoch=False, average=True, **kwargs):
        """
        Average all weights over all updates, as a form of regularization
        :param average: whether to really average the weights or just return them as they are now
        :param finished_epoch: whether to decay the learning rate and drop rare features
        :return new SparsePerceptron object with the weights averaged
        """
        super().finalize(finished_epoch=finished_epoch, **kwargs)
        started = time.time()
        if average:
            print("Averaging weights... ", end="", flush=True)
        finalized = self._finalize_model(finished_epoch, average)
        if average:
            print("Done (%.3fs)." % (time.time() - started))
        print(self)
        return finalized

    def _finalize_model(self, finished_epoch, average):
        # If finished an epoch, remove rare features from our model directly. Otherwise, copy it.
        model, dropped = (self.model, self.dropped) if finished_epoch else (self.copy_model(), set())
        num_features = 0
        for axis, axis_model in self.model.items():
            for feature, weights in list(axis_model.items()):
                if weights.update_count < self.min_update:
                    del axis_model[feature]
                    dropped.add(feature)
            num_features += len(axis_model)
        print("%d features occurred at least %d times, dropped %d rare features" % (
            num_features, self.min_update, len(dropped)))
        finalized = {a: {f: w.finalize(self.updates, average=average) for f, w in m.items()} for a, m in model.items()}
        ret = SparsePerceptron(self.config, self.labels, epoch=self.epoch)
        ret.update_model(finalized)
        ret.is_frozen = True
        ret.min_update = self.min_update
        ret.updates = self.updates
        ret.initial_learning_rate = self.initial_learning_rate
        ret.learning_rate = self.learning_rate
        ret.learning_rate_decay = self.learning_rate_decay
        ret.epoch = self.epoch
        return ret

    def save_model(self, filename, d):
        super().save_model(filename, d)
        d.update((
            ("initial_learning_rate", self.initial_learning_rate),
            ("min_update", self.min_update),
        ))
        save_dict(filename + ".data", self.copy_model())

    def load_model(self, filename, d):
        self.model.clear()
        self.update_model(load_dict(filename + ".data"))
        self.initial_learning_rate = d["initial_learning_rate"]
        self.config.args.min_update = self.min_update = d["min_update"]
        super().load_model(filename, d)

    def all_params(self):
        d = super().all_params()
        d.update(("_".join((axis, k)), v.weights) for axis, model in self.model.items() for k, v in model.items())
        return d

    def print_params(self, max_rows=10):
        for axis, model in self.model.items():
            for key, value in model.params.items():
                print(axis, key)
                # noinspection PyBroadException
                try:
                    print(list(value.items())[:max_rows])
                except Exception:
                    pass
